import copy

from tala.model import move
from tala.model.entity import Entity
from tala.model import stack
from tala.model import plan_item
from tala.model import ontology
from tala.model.domain import Domain
from tala.model import condition
from tala.model.hint import Hint
from tala.model.action import Action
from tala.model.ask_feature import AskFeature
from tala.model.plan import Plan
from tala.model import sort
from tala.model.sort import Sort, CustomSort
from tala.model.predicate import Predicate
from tala.model.individual import No, Yes, Individual, NegativeIndividual
from tala.model.question import Question, WhQuestion, YesNoQuestion, ConsequentQuestion, AltQuestion, \
    KnowledgePreconditionQuestion
from tala.model.lambda_abstraction import LambdaAbstractedProposition, \
    LambdaAbstractedImplicationPropositionForConsequent, LambdaAbstractedGoalProposition, \
    LambdaAbstractedPredicateProposition
from tala.model.proposition import Proposition, GoalProposition, PropositionSet, PredicateProposition, \
    ResolvednessProposition, PrereportProposition, ImplicationProposition, UnderstandingProposition, \
    KnowledgePreconditionProposition, PreconfirmationProposition, RejectedPropositions, ServiceResultProposition, \
    ActionStatusProposition, ServiceActionStartedProposition, ServiceActionTerminatedProposition, \
    NumberOfAlternativesProposition, QuestionStatusProposition, MuteProposition, UnmuteProposition, QuitProposition
from tala.model.polarity import Polarity
from tala.model.goal import Goal, PerformGoal, ResolveGoal, HandleGoal
from tala.model.set import Set
from tala.model.openqueue import OpenQueue
from tala.model.date_time import DateTime
from tala.model.person_name import PersonName
from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction
from tala.model.action_status import ActionStatus, Aborted, Done
from tala.model.ddd import DDD
from tala.ddd.services import service_interface
from tala.ddd.services.service_interface import ServiceInterface, ServiceActionInterface


class JSONParseFailure(Exception):
    pass


def preprocess_json_dict(json_dict_original):
    json_dict = copy.copy(json_dict_original)
    for float_key in ["perception_confidence", "understanding_confidence"]:
        if float_key in json_dict:
            try:
                json_dict[float_key] = float(json_dict[float_key])
            except TypeError:
                pass
    return json_dict


def create_sort(definition):
    sort_name = definition["_name"]
    if sort_name in sort.BUILTIN_SORTS:
        if sort_name == sort.DATETIME:
            return sort.DateTimeSort()
        if sort_name == sort.REAL:
            return sort.RealSort()
        if sort_name == sort.PERSON_NAME:
            return sort.PersonNameSort()
        if sort_name == sort.INTEGER:
            return sort.IntegerSort()
        if sort_name == sort.BOOLEAN:
            return sort.BooleanSort()
        if sort_name == sort.STRING:
            return sort.StringSort()
        if sort_name == sort.DOMAIN:
            return sort.DomainSort()
        if sort_name == sort.IMAGE:
            return sort.ImageSort()
        if sort_name == sort.WEBVIEW:
            return sort.WebviewSort()

    ontology_name = definition.get("_ontology_name")
    if ontology_name:
        return CustomSort(ontology_name, definition["_name"], definition["_dynamic"])
    return Sort(definition["_name"], definition["_dynamic"])


def create_predicate(name, definition):
    return Predicate(
        definition["_ontology_name"], name, create_sort(definition["sort"]), definition["feature_of_name"],
        definition["_multiple_instances"]
    )


def create_individual(individual, sort):
    def extract(string, prefix, suffix):
        if value.startswith(prefix) and value.endswith(suffix):
            return value[len(prefix):len(value) - len(suffix)]

    def extract_datetime_value():
        return extract(value, "datetime(", ")")

    def extract_person_name_value():
        return extract(value, "person_name(", ")")

    sort = create_sort(sort)
    value = individual
    if sort.is_datetime_sort():
        value = DateTime(extract_datetime_value())
    if sort.is_real_sort():
        value = float(value)
    if sort.is_person_name_sort():
        value = PersonName(extract_person_name_value())
    if sort.is_integer_sort():
        value = int(value)
    return (value, sort)


class JSONOntologyParser():
    def parse_ontology(self, ontology_as_json):
        if ontology_as_json["semantic_object_type"] == ontology.SEMANTIC_OBJECT_TYPE:
            name = ontology_as_json["_name"]
            sorts = self.parse_ontology_sorts(ontology_as_json["_sorts"])
            predicates = self.parse_ontology_predicates(ontology_as_json["_predicates"])
            individuals = self.parse_ontology_individuals(ontology_as_json["_individuals"])
            actions = self.parse_ontology_actions(ontology_as_json["_actions"], name)
            return ontology.Ontology(name, sorts, predicates, individuals, actions)
        raise JSONParseFailure(f"Failed to parse {ontology_as_json} as an ontology.")

    def parse_ontology_sorts(self, sorts_as_json):
        return set([create_sort(sort) for sort in sorts_as_json])

    def parse_ontology_predicates(self, predicates_as_json):
        return set([create_predicate(name, predicates_as_json[name]) for name in predicates_as_json])

    def parse_ontology_individuals(self, individuals_as_json):
        individuals = {}
        for individual_as_json in individuals_as_json:
            (value, sort) = create_individual(individual_as_json, individuals_as_json[individual_as_json])
            individuals[value] = sort
        return individuals

    def parse_ontology_actions(self, actions_as_json, ontology_name):
        return [action for action in actions_as_json["set"]]


class JSONDDDParser():
    def parse(self, ddd_as_json):
        name = ddd_as_json["ddd_name"]
        ontology_parser = JSONOntologyParser()
        ontology = ontology_parser.parse_ontology(ddd_as_json["ontology"])
        domain_parser = JSONDomainParser(ontology)
        domain = domain_parser.parse(ddd_as_json["domain"])
        service_interface_parser = JSONServiceInterfaceParser()
        service_interface = service_interface_parser.parse(ddd_as_json["service_interface"])
        return DDD(name, ontology, domain, service_interface)


class JSONDomainParser():
    def __init__(self, ontology=None):
        self.ontology = ontology
        self.parser = NonCheckingJSONParser()

    def parse(self, json_dict):
        ddd_name = json_dict["ddd_name"]
        domain_name = json_dict["name"]
        if self.ontology:
            self.parser = CheckingJSONParser(ddd_name, self.ontology, domain_name)
        default_questions = self.parse_default_questions(json_dict["default_questions"])
        parameters = self.parse_parameters(json_dict["parameters"])
        plans = self.parse_plans(json_dict["plans"])
        validators = self.parse_validators(json_dict["validators"])
        dependencies = self.parse_serialized_dict(json_dict.get("dependencies", {}))
        queries = self.parse_serialized_query_list(json_dict["queries"])
        return Domain(
            ddd_name, domain_name, self.ontology, plans, default_questions, parameters, validators, dependencies,
            queries
        )

    def parse_default_questions(self, default_questions):
        return [self._parse_default_question(question) for question in default_questions]

    def _parse_default_question(self, question):
        return self.parser.parse(question)

    def parse_parameters(self, parameters):
        compiled_parameters = {}
        for pair in parameters:
            key_question = self.parser.parse(pair["key"])
            values = pair["value"]
            compiled_parameters[key_question] = self._parse_parameters_for_entry(values)
        return compiled_parameters

    def _parse_parameters_for_entry(self, parameters):
        return {name: self._parse_parameter(parameters, name) for name in parameters}

    def _parse_parameter(self, parameters, name):
        def is_questions_parameter():
            return name in ["related_information"]

        def is_question_parameter():
            return name in ["service_query", "default"]

        def is_int_parameter():
            return name in ["max_spoken_alts", "max_reported_hit_count"]

        def is_boolean_parameter():
            return name in ["always_ground", "always_relevant"]

        if name == "ask_features":
            return [self.parse_ask_feature(feature) for feature in parameters[name]]
        if name == "hints":
            return [self.parse_hint(hint) for hint in parameters[name]]
        if name in ["allow_goal_accommodation", "verbalize", "incremental"]:
            return parameters[name]
        if name in ["on_zero_hits_action", "on_too_many_hits_action"]:
            return self.parser.parse_action(parameters[name])
        if is_int_parameter():
            return int(parameters[name])
        if name in ["graphical_type", "source"]:
            return parameters[name]
        if name == "alts":
            return PropositionSet([self.parser.parse(proposition) for proposition in parameters[name]["_propositions"]])
        if name == "background":
            return [self.parser.parse_predicate(predicate) for predicate in parameters[name]]
        if is_questions_parameter():
            return [self.parser.parse_question(question) for question in parameters[name]]
        if is_question_parameter():
            return self.parser.parse_question(parameters[name])
        if is_boolean_parameter():
            return bool(parameters[name])

        raise JSONParseFailure("no support for parameter", name, parameters[name], ".")
        return parameters[name]

    def parse_ask_feature(self, feature):
        return AskFeature(feature["name"], feature["kpq"])

    def parse_hint(self, hint):
        return Hint([self.parser.parse_plan_item(item) for item in hint["_plan"]])

    def parse_plans(self, plans):
        def is_builtin(goal):
            return goal["_goal_type"] == "PERFORM_GOAL" and goal["_content"]["value"] in ["top", "up"]

        return [self.parse_goal(plans[goal]) for goal in plans]

    def parse_goal(self, goal):
        result = {}
        BOOLEAN_VALUES = [
            "unrestricted_accommodation", "restart_on_completion", "reraise_on_resume", "accommodate_without_feedback"
        ]
        INTEGER_VALUES = ["max_answers"]
        STRING_VALUES = ["alternatives_predicate"]
        for entry in goal:
            if entry == "goal":
                result[entry] = self.parser.parse_goal(goal[entry])
            elif entry == "io_status":
                result[entry] = self.parse_io_status(goal[entry])
            elif entry == "plan":
                result[entry] = self.parse_plan(goal[entry])
            elif entry == "postconds":
                result[entry] = self.parse_postconds(goal[entry])
            elif entry == "postplan":
                result[entry] = self.parse_postplan(goal[entry])
            elif entry == "preferred":
                result[entry] = self.parse_preferred(goal[entry])
            elif entry == "superactions":
                result[entry] = self.parse_superactions(goal[entry])
            elif entry in BOOLEAN_VALUES:
                result[entry] = bool(goal[entry])
            elif entry in INTEGER_VALUES:
                result[entry] = int(goal[entry])
            elif entry in STRING_VALUES:
                result[entry] = str(goal[entry])
            else:
                raise JSONParseFailure("found unknown entry in goal when parsing:", entry, goal[entry])

        return result

    def parse_plan(self, plan):
        plan_items = [self.parser.parse_plan_item(item) for item in plan["stack"]]
        return Plan(reversed(plan_items))

    def parse_io_status(self, status):
        if status:
            return status
        return Domain.DEFAULT_IO_STATUS

    def parse_postconds(self, postconds):
        return [self.parse_postcond(postcond) for postcond in postconds]

    def parse_postcond(self, postcond):
        return self.parser.parse(postcond)

    def parse_postplan(self, postplan):
        try:
            return [self.parser.parse_plan_item(item) for item in postplan["stack"]]
        except TypeError:
            return [self.parser.parse_plan_item(item) for item in postplan]

    def parse_superactions(self, actions):
        return [self.parse_superaction(action) for action in actions]

    def parse_preferred(self, preference):
        try:
            return self.parser.parse_proposition(preference)
        except TypeError:
            return preference

    def parse_superaction(self, action):
        return self.parser.parse_action(action)

    def parse_validators(self, validators):
        return [self.parse_validator(validator) for validator in validators]

    def parse_validator(self, validator):
        def parse_propositions_list(predicates):
            return [self.parser.parse(predicate) for predicate in predicates]

        id = validator["validator"]
        valid_configurations = [
            parse_propositions_list(configuration) for configuration in validator["valid_configurations"]
        ]
        return {"validator": id, "valid_configurations": valid_configurations}

    def parse_serialized_dict(self, serialized_dict):
        result_dict = {}
        for pair in serialized_dict:
            key = self.parser.parse(pair["key"])
            value = pair["value"]
            result_dict[key] = self.parser.parse(value)
        return result_dict

    def parse_serialized_query_list(self, serialized_queries):
        def is_iterator(query_content):
            return "limit" in query_content

        def parse_iterator(iterator):
            result_dict = {}
            for key, value in iterator.items():
                if key in ["for_enumeration", "for_random_enumeration"]:
                    result_dict[key] = [self.parser.parse(item) for item in value]
                else:
                    result_dict[key] = value
            return result_dict

        result_list = []
        for query in serialized_queries:
            if is_iterator(query):
                query_dict = parse_iterator(query)
            else:
                query_dict = {}
                for entry, semantic_objects in query.items():
                    try:
                        query_dict[entry] = self.parser.parse(semantic_objects)
                    except AttributeError:
                        query_dict[entry] = [self.parser.parse(item) for item in semantic_objects]
            result_list.append(query_dict)
        return result_list


class JSONServiceInterfaceParser():
    def parse(self, json_dict):
        actions = self.parse_actions(json_dict["_actions"])
        queries = self.parse_queries(json_dict["_queries"])
        validators = self.parse_validators(json_dict["_validators"])
        return ServiceInterface(actions, queries, validators)

    def parse_actions(self, actions_dict):
        return [self.parse_action_interface(actions_dict[action_name]) for action_name in actions_dict]

    def parse_action_interface(self, action_dict):
        name = action_dict["_name"]
        target = self.parse_target(action_dict["_target"])
        service_parameters = self.parse_parameters(action_dict["_parameters"])
        failure_reasons = action_dict["_failure_reasons"]
        if action_dict.get("_audio_url_parameter"):
            audio_parameter = self.parse_audio_parameter(action_dict["_audio_url_parameter"])
            interface = service_interface.PlayAudioActionInterface(name, target, service_parameters, audio_parameter)
        else:
            interface = ServiceActionInterface(name, target, service_parameters, failure_reasons)
        return interface

    def parse_target(self, target_dict):
        if target_dict.get("target_type") == service_interface.DEVICE_MODULE_TARGET:
            return service_interface.DeviceModuleTarget(target_dict["_device"])
        if target_dict.get("target_type") == service_interface.FRONTEND_TARGET:
            return service_interface.FrontendTarget()
        if target_dict.get("target_type") == service_interface.HTTP_TARGET:
            endpoint = target_dict["_endpoint"]
            return service_interface.HttpTarget(endpoint)
        raise JSONParseFailure(f"Cannot parse as target: {target_dict}")

    def parse_parameters(self, parameter_dict):
        return [self.parse_parameter(parameter) for parameter in parameter_dict if parameter["_name"] != "audio_url"]

    def parse_parameter(self, parameter):
        name = parameter["_name"]
        format = parameter["_format"]
        is_optional = parameter["_is_optional"]
        return service_interface.ServiceParameter(name, format, is_optional)

    def parse_audio_parameter(self, parameter):
        return service_interface.AudioURLServiceParameter(parameter["_name"])

    def parse_queries(self, queries_dict):
        return [self.parse_query(query) for query in queries_dict.values()]

    def parse_query(self, query_dict):
        name = query_dict["_name"]
        target = self.parse_target(query_dict["_target"])
        parameters = self.parse_parameters(query_dict["_parameters"])
        return service_interface.ServiceQueryInterface(name, target, parameters)

    def parse_validators(self, validators_dict):
        return [self.parse_validator(validator) for validator in validators_dict.values()]

    def parse_validator(self, validator_dict):
        name = validator_dict["_name"]
        target = self.parse_target(validator_dict["_target"])
        parameters = self.parse_parameters(validator_dict["_parameters"])
        return service_interface.ServiceValidatorInterface(name, target, parameters)


class NonCheckingJSONParser():
    def parse(self, json_dict_unchecked):
        def is_condition():
            """ This is the template for extending the semantic object JSON
            with type information, which should be done for all semantic objects.
            """

            return json_dict.get("semantic_object_type") == condition.SEMANTIC_OBJECT_TYPE

        def is_plan_item():
            for item in plan_item.ALL_PLAN_ITEM_TYPES:
                if item in json_dict:
                    return True
            return False

        def is_individual():
            return "value" in json_dict and "sort" in json_dict

        def is_yes_no():
            return json_dict.get("semantic_object_type") == "yes/no"

        def is_move():
            return "move_type" in json_dict

        def is_action():
            return "value" in json_dict

        def is_goal():
            return "_goal_type" in json_dict

        def is_question():
            question_type = json_dict.get("_type", None)
            return question_type in Question.TYPES

        def is_predicate():
            return "name" in json_dict and "sort" in json_dict

        def is_lambda_abstracted_proposition():
            return json_dict.get("semantic_object_type") == "LambdaAbstractedProposition"

        def is_set():
            return "set" in json_dict

        def is_openqueue():
            return "openqueue" in json_dict

        def is_service_action_outcome():
            return "service_action_outcome" in json_dict

        def is_proposition():
            prop_type = json_dict.get("_type", None)
            return prop_type in Proposition.TYPES

        def is_proposition_set():
            return "_propositions" in json_dict

        def is_action_status():
            return "action_status" in json_dict

        json_dict = preprocess_json_dict(json_dict_unchecked)
        if is_condition():
            return self.parse_condition(json_dict)
        if is_plan_item():
            return self.parse_plan_item(json_dict)
        if is_individual():
            return self.parse_individual(json_dict)
        if is_set():
            return self.parse_set(json_dict["set"])
        if is_openqueue():
            return self.parse_openqueue(json_dict["openqueue"])
        if is_move():
            return self.parse_move(json_dict)
        if is_action():
            return self.parse_action(json_dict)
        if is_goal():
            return self.parse_goal(json_dict)
        if is_question():
            return self.parse_question(json_dict)
        if is_predicate():
            return self.parse_predicate(json_dict)
        if is_proposition_set():
            return self.parse_proposition_set(json_dict)
        if is_proposition():
            return self.parse_proposition(json_dict)
        if is_lambda_abstracted_proposition():
            return self.parse_lambda_abstraction(json_dict)
        if is_service_action_outcome():
            return self.parse_service_action_outcome(json_dict)
        if is_yes_no():
            if json_dict["instance"] == No.NO:
                return No()
            if json_dict["instance"] == Yes.YES:
                return Yes()
        if is_action_status():
            return self.parse_action_status(json_dict)
        raise JSONParseFailure(f"JSONParser cannot parse: {json_dict}")

    def parse_condition(self, condition_json):
        keys = ["_predicate", "_proposition"]
        for key in keys:
            if condition_json.get(key):
                argument = self.parse(condition_json[key])
                return condition.create_condition(condition_json["type"], argument)
        if condition_json.get("_query"):
            try:
                argument = self.parse(condition_json["_query"])
            except Exception:
                argument = condition_json["_query"]
            return condition.create_condition(condition_json["type"], argument)
        raise JSONParseFailure(f"Failed to parse {condition_json} as a condition.")

    def parse_plan_item(self, data):
        if plan_item.TYPE_FINDOUT in data:
            return self.parse_findout_plan_item(data)
        if plan_item.TYPE_RAISE in data:
            return self.parse_raise_plan_item(data)
        if plan_item.TYPE_RESPOND in data:
            return self.parse_respond_plan_item(data[plan_item.TYPE_RESPOND])
        if plan_item.TYPE_ASSUME in data:
            return self.parse_assume_plan_item(data[plan_item.TYPE_ASSUME])
        if plan_item.TYPE_ASSUME_ISSUE in data:
            return self.parse_assume_issue_plan_item(data)
        if plan_item.TYPE_ASSUME_SHARED in data:
            return self.parse_assume_shared_plan_item(data[plan_item.TYPE_ASSUME_SHARED])
        if plan_item.TYPE_BIND in data:
            return self.parse_bind_plan_item(data[plan_item.TYPE_BIND])
        if plan_item.TYPE_DO in data:
            return self.parse_do_plan_item(data[plan_item.TYPE_DO])
        if plan_item.TYPE_EMIT_ICM in data:
            return self.parse_emit_icm_plan_item(data[plan_item.TYPE_EMIT_ICM])
        if plan_item.TYPE_FORGET in data:
            return self.parse_forget_plan_item(data[plan_item.TYPE_FORGET])
        if plan_item.TYPE_FORGET_SHARED in data:
            return self.parse_forget_shared_plan_item(data[plan_item.TYPE_FORGET_SHARED])
        if plan_item.TYPE_FORGET_ALL in data:
            return self.parse_forget_all_plan_item(data[plan_item.TYPE_FORGET_ALL])
        if plan_item.TYPE_FORGET_ISSUE in data:
            return self.parse_forget_issue_plan_item(data[plan_item.TYPE_FORGET_ISSUE])
        if plan_item.TYPE_IF_THEN_ELSE in data:
            return self.parse_if_then_else_plan_item(data[plan_item.TYPE_IF_THEN_ELSE])
        if plan_item.TYPE_JUMPTO in data:
            return self.parse_jumpto_plan_item(data[plan_item.TYPE_JUMPTO])
        if plan_item.TYPE_INVOKE_SERVICE_QUERY in data:
            return self.parse_invoke_service_query_plan_item(data[plan_item.TYPE_INVOKE_SERVICE_QUERY])
        if plan_item.TYPE_INVOKE_DOMAIN_QUERY in data:
            return self.parse_invoke_domain_query_plan_item(data[plan_item.TYPE_INVOKE_DOMAIN_QUERY])
        if plan_item.TYPE_INVOKE_SERVICE_ACTION in data:
            return self.parse_invoke_service_action_plan_item(data[plan_item.TYPE_INVOKE_SERVICE_ACTION])
        if plan_item.TYPE_LOG in data:
            return self.parse_log_plan_item(data[plan_item.TYPE_LOG])
        if plan_item.TYPE_ACTION_PERFORMED in data:
            return self.parse_action_performed_plan_item(data[plan_item.TYPE_ACTION_PERFORMED])
        if plan_item.TYPE_ACTION_ABORTED in data:
            return self.parse_action_aborted_plan_item(data[plan_item.TYPE_ACTION_ABORTED])
        if plan_item.TYPE_CHANGE_DDD in data:
            return self.parse_change_ddd_plan_item(data[plan_item.TYPE_CHANGE_DDD])
        if plan_item.TYPE_GET_DONE in data:
            return self.parse_get_done_plan_item(data)
        if plan_item.TYPE_INVOKE_DOMAIN_QUERY in data:
            return self.parse_invoke_domain_query(data[plan_item.INVOKE_DOMAIN_QUERY])
        if plan_item.TYPE_END_TURN in data:
            return self.parse_end_turn_plan_item(data[plan_item.TYPE_END_TURN])
        if plan_item.TYPE_RESET_DOMAIN_QUERY in data:
            return self.parse_reset_domain_query(data[plan_item.TYPE_RESET_DOMAIN_QUERY])
        if plan_item.TYPE_ITERATE in data:
            return self.parse_iterate_plan_item(data[plan_item.TYPE_ITERATE])
        if plan_item.TYPE_ACTION_REPORT in data:
            return self.parse_action_report_plan_item(data[plan_item.TYPE_ACTION_REPORT])
        if plan_item.TYPE_GREET in data:
            return plan_item.Greet()
        if plan_item.TYPE_EMIT_MOVE in data:
            return self.parse_emit_move_plan_item(data)
        if plan_item.TYPE_QUESTION_REPORT in data:
            return self.parse_question_report_plan_item(data)
        if plan_item.TYPE_SERVICE_REPORT in data:
            return self.parse_service_report_plan_item(data)
        if plan_item.TYPE_RESPOND_TO_INSULT in data:
            return plan_item.RespondToInsult()
        if plan_item.TYPE_RESPOND_TO_THANK_YOU in data:
            return plan_item.RespondToThankYou()

        raise JSONParseFailure(f"PlanItem {data} not supported by json parser")

    def parse_findout_plan_item(self, data):
        allow_answer_from_pcom = data.get("allow_answer_from_pcom", False)
        domain_name = data.get("domain_name")
        question = self.parse_question(data[plan_item.TYPE_FINDOUT])
        return plan_item.Findout(domain_name, question, allow_answer_from_pcom)

    def parse_raise_plan_item(self, data):
        domain_name = data.get("domain_name")
        question = self.parse_question(data[plan_item.TYPE_RAISE])
        return plan_item.Raise(domain_name, question)

    def parse_respond_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.Respond(question)

    def parse_bind_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.Bind(question)

    def parse_do_plan_item(self, data):
        action = self.parse_action(data)
        return plan_item.Do(action)

    def parse_log_plan_item(self, data):
        message = data
        return plan_item.Log(message)

    def parse_action_performed_plan_item(self, data):
        return plan_item.GoalPerformed(data)

    def parse_action_aborted_plan_item(self, data):
        return plan_item.GoalAborted(data)

    def parse_change_ddd_plan_item(self, data):
        return plan_item.ChangeDDD(data)

    def parse_get_done_plan_item(self, data):
        action = self.parse_action(data[plan_item.TYPE_GET_DONE])
        try:
            step = int(data.get("step"))
        except TypeError:
            step = None

        return plan_item.GetDone(action, step)

    def parse_end_turn_plan_item(self, data):
        timeout = float(data)
        return plan_item.EndTurn(timeout)

    def parse_reset_domain_query(self, data):
        query = self.parse_question(data)
        return plan_item.ResetDomainQuery(query)

    def parse_iterate_plan_item(self, iterator_name):
        return plan_item.Iterate(iterator_name)

    def parse_action_report_plan_item(self, result_proposition):
        return plan_item.ActionReport(self.parse_proposition(result_proposition))

    def parse_emit_move_plan_item(self, data):
        prereport_move = self.parse_prereport_move(data[plan_item.TYPE_EMIT_MOVE])
        return plan_item.EmitMove(prereport_move)

    def parse_invoke_domain_query(self, data):
        issue = self.parse_question(data)
        return plan_item.InvokeDomainQuery(issue)

    def parse_forget_plan_item(self, data):
        if "_type" in data.keys():
            proposition = self.parse_proposition(data)
            return plan_item.Forget(proposition)
        else:
            predicate = self.parse_predicate(data)
            return plan_item.Forget(predicate)

    def parse_forget_shared_plan_item(self, data):
        if "_type" in data.keys():
            proposition = self.parse_proposition(data)
            return plan_item.ForgetShared(proposition)
        else:
            predicate = self.parse_predicate(data)
            return plan_item.ForgetShared(predicate)

    def parse_forget_all_plan_item(self, _data):
        return plan_item.ForgetAll()

    def parse_forget_issue_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.ForgetIssue(question)

    def parse_assume_plan_item(self, data):
        proposition = self.parse_proposition(data)
        return plan_item.Assume(proposition)

    def parse_assume_issue_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_ASSUME_ISSUE])
        return plan_item.AssumeIssue(question, data["insist"])

    def parse_assume_shared_plan_item(self, data):
        proposition = self.parse_proposition(data)
        return plan_item.AssumeShared(proposition)

    def parse_emit_icm_plan_item(self, data):
        icm = self.parse_move(data)
        return plan_item.EmitIcm(icm)

    def parse_jumpto_plan_item(self, data):
        goal = self.parse_goal(data)
        return plan_item.JumpTo(goal)

    def parse_if_then_else_plan_item(self, data):
        cond = self.parse(data["condition"])
        consequent = [self.parse(element) for element in data["consequent"]]
        alternative = [self.parse(element) for element in data["alternative"]]
        return plan_item.IfThenElse(cond, consequent, alternative)

    def parse_invoke_service_query_plan_item(self, data):
        issue = self.parse_question(data["issue"])
        min_results = int(data["min_results"])
        max_results = int(data["max_results"])
        return plan_item.InvokeServiceQuery(issue, min_results=min_results, max_results=max_results)

    def parse_invoke_domain_query_plan_item(self, data):
        issue = self.parse_question(data["issue"])
        min_results = int(data["min_results"])
        max_results = int(data["max_results"])
        return plan_item.InvokeDomainQuery(issue, min_results=min_results, max_results=max_results)

    def parse_invoke_service_action_plan_item(self, data):
        service_action = data["service_action"]
        preconfirm = data["preconfirm"]
        postconfirm = data["postconfirm"]
        downdate_plan = data["downdate_plan"]
        ontology_name = data["ontology"]
        return plan_item.InvokeServiceAction(ontology_name, service_action, preconfirm, postconfirm, downdate_plan)

    def parse_question_report_plan_item(self, data):
        result_proposition = self.parse_proposition(data[plan_item.TYPE_QUESTION_REPORT])
        return plan_item.QuestionReport(result_proposition)

    def parse_service_report_plan_item(self, data):
        result_proposition = self.parse_proposition(data[plan_item.TYPE_SERVICE_REPORT])
        return plan_item.ServiceReport(result_proposition)

    def parse_question(self, data):
        if data["_type"] == Question.TYPE_WH:
            lambda_abstraction = self.parse_lambda_abstraction(data["_content"])
            return WhQuestion(lambda_abstraction)
        if data["_type"] == Question.TYPE_KPQ:
            embedded_question = self.parse_question(data["_content"])
            return KnowledgePreconditionQuestion(embedded_question)
        if data["_type"] == Question.TYPE_YESNO:
            proposition = self.parse(data["_content"])
            return YesNoQuestion(proposition)
        if data["_type"] == Question.TYPE_ALT:
            p_set = self.parse_proposition_set(data["_content"])
            return AltQuestion(p_set)
        if data["_type"] == Question.TYPE_CONSEQUENT:
            lambda_abstraction = self.parse_lambda_abstraction(data["_content"])
            return ConsequentQuestion(lambda_abstraction)
        raise JSONParseFailure(f"Cannot parse {data} as question.")

    def parse_lambda_abstraction(self, content):
        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_PREDICATE_PROPOSITION:
            return self._create_lambda_abstracted_predicate_proposition(content)

        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_GOAL_PROPOSITION:
            return LambdaAbstractedGoalProposition()

        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_IMPLICATION_PROPOSITION_FOR_CONSEQUENT:
            ontology_name = content["_ontology_name"]
            antecedent = self.parse_proposition(content["_antecedent"])
            predicate = self.parse_predicate(content["_consequent_predicate"])
            return LambdaAbstractedImplicationPropositionForConsequent(antecedent, predicate, ontology_name)
        raise JSONParseFailure(f"Cannot parse {content} as lambda abstraction.")

    def _create_lambda_abstracted_predicate_proposition(self, content):
        predicate = self.parse_predicate(content["predicate"])
        ontology_name = content["_ontology_name"]
        return LambdaAbstractedPredicateProposition(predicate, ontology_name)

    def parse_predicate(self, data):
        return create_predicate(data["name"], data)

    def parse_proposition_set(self, data):
        proposition_list = data["_propositions"]
        propositions = [self.parse_proposition(proposition) for proposition in proposition_list]
        polarity = data["_polarity"]
        return PropositionSet(propositions, polarity)

    def parse_set(self, data):
        result_set = Set([self.parse(element) for element in data])
        return result_set

    def parse_openqueue(self, input):
        front_content = [self.parse(item) for item in input["front_content"]]
        back_content = [self.parse(item) for item in input["back_content"]]
        unshifted_as_json = input["unshifted_content"]
        if unshifted_as_json:
            unshifted_content = [self.parse(item) for item in unshifted_as_json]
        else:
            unshifted_content = None
        return OpenQueue(front_content, back_content, unshifted_content)

    def parse_move(self, data):
        def get_float(string):
            if string is not None:
                return float(string)
            return None

        move_args = {
            "understanding_confidence": get_float(data.get("understanding_confidence")),
            "speaker": data.get("speaker"),
            "ddd_name": data.get("ddd"),
            "perception_confidence": get_float(data.get("perception_confidence"))
        }
        non_icm_args = {
            "modality": data.get("modality"),
            "weighted_understanding_confidence": get_float(data.get("weighted_understanding_confidence")),
            "utterance": data.get("utterance")
        } | move_args

        icm_args = {
            "polarity": data.get("polarity"),
        } | move_args

        if data["move_type"] == move.Move.ANSWER:
            return self.parse_answer_move(data, non_icm_args)
        if data["move_type"] == move.Move.REQUEST:
            return self.parse_request_move(data, non_icm_args)
        if data["move_type"] == move.Move.ASK:
            return self.parse_ask_move(data, non_icm_args)
        if data["move_type"] == move.ICMMove.ACC:
            return self.parse_acc_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.ACCOMMODATE:
            return self.parse_accommodate_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.RESUME:
            return self.parse_resume_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.RERAISE:
            return self.parse_reraise_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.PER:
            return self.parse_per_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.UND:
            return self.parse_und_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.SEM:
            return self.parse_sem_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.LOADPLAN:
            return self.parse_loadplan_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.CARDINAL_SEQUENCING:
            return self.parse_cardinal_sequencing_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.REPORT_INFERENCE:
            return self.parse_report_inference_icm(data, icm_args)
        if data["move_type"] == move.Move.GREET:
            return self.parse_greet_move(non_icm_args)
        if data["move_type"] == move.Move.INSULT:
            return self.parse_insult_move(non_icm_args)
        if data["move_type"] == move.Move.INSULT_RESPONSE:
            return self.parse_insult_response_move(non_icm_args)
        if data["move_type"] == move.Move.MUTE:
            return self.parse_mute_move(non_icm_args)
        if data["move_type"] == move.Move.UNMUTE:
            return self.parse_unmute_move(non_icm_args)
        if data["move_type"] == move.Move.PREREPORT:
            return self.parse_prereport_move(data)
        if data["move_type"] == move.Move.REPORT:
            return self.parse_report_move(data, non_icm_args)
        if data["move_type"] == move.Move.QUIT:
            return self.parse_quit_move(non_icm_args)
        if data["move_type"] == move.Move.THANK_YOU:
            return self.parse_thank_you_move(non_icm_args)
        if data["move_type"] == move.Move.THANK_YOU_RESPONSE:
            return self.parse_thank_you_response_move(non_icm_args)
        raise JSONParseFailure(f"cannot parse {data} as move.")

    def parse_action(self, data):
        action_name = data["value"]
        ontology_name = data["_ontology_name"]
        return Action(action_name, ontology_name)

    def parse_answer_move(self, data, kwargs):
        content = self.parse(data["content"])
        return move.AnswerMove(content, **kwargs)

    def parse_request_move(self, data, kwargs):
        action = self.parse_action(data["content"])
        return move.RequestMove(action, **kwargs)

    def parse_proposition(self, data):
        if data["_type"] == Proposition.PREDICATE:
            return self._create_predicate_proposition(data)
        if data["_type"] == Proposition.GOAL:
            goal = self.parse_goal(data["_content"])
            polarity = data.get("_polarity")
            return GoalProposition(goal, polarity)
        if data["_type"] == Proposition.PRECONFIRMATION:
            service_action = data.get("service_action")
            polarity = data.get("polarity")
            arguments = [self.parse_proposition(proposition) for proposition in data.get("_arguments")]
            ontology_name = data["_ontology_name"]
            return PreconfirmationProposition(ontology_name, service_action, arguments, polarity)
        if data["_type"] == Proposition.SERVICE_RESULT:
            service_action = data.get("service_action")
            action_outcome = self.parse(data.get("result"))
            arguments = [self.parse_proposition(proposition) for proposition in data.get("arguments")]
            ontology_name = data.get("ontology_name")
            return ServiceResultProposition(ontology_name, service_action, arguments, action_outcome)
        if data["_type"] == Proposition.UNDERSTANDING:
            content = self.parse(data["_content"])
            polarity = data.get("_polarity")
            speaker = data.get("_speaker")
            return UnderstandingProposition(speaker, content, polarity)
        if data["_type"] == Proposition.RESOLVEDNESS:
            issue = self.parse_question(data["issue"])
            return ResolvednessProposition(issue)
        if data["_type"] == Proposition.KNOWLEDGE_PRECONDITION:
            question = self.parse_question(data["_content"])
            polarity = data.get("_polarity")
            return KnowledgePreconditionProposition(question, polarity)
        if data["_type"] == Proposition.ACTION_STATUS:
            status = self.parse_action_status(data["_status"])
            action = self.parse_action(data["_content"])
            return ActionStatusProposition(action, status)
        if data["_type"] == Proposition.SERVICE_ACTION_STARTED:
            service_action = data["service_action"]
            parameters = [self.parse_proposition(proposition) for proposition in data.get("parameters")]
            ontology_name = data.get("ontology_name")
            return ServiceActionStartedProposition(ontology_name, service_action, parameters)
        if data["_type"] == Proposition.SERVICE_ACTION_TERMINATED:
            service_action = data["service_action"]
            polarity = data.get("_polarity")
            ontology_name = data["_ontology_name"]
            return ServiceActionTerminatedProposition(ontology_name, service_action, polarity)
        if data["_type"] == Proposition.PREREPORT:
            service_action = data["service_action"]
            argument_list = [self.parse_proposition(proposition) for proposition in data["argument_set"]]
            ontology_name = data["_ontology_name"]
            return PrereportProposition(ontology_name, service_action, argument_list)
        if data["_type"] == Proposition.REJECTED:
            polarity = data.get("_polarity")
            reason_for_rejection = data.get("reason_for_rejection")
            combination = self.parse(data["rejected_combination"])
            return RejectedPropositions(combination, polarity, reason_for_rejection)
        if data["_type"] == Proposition.IMPLICATION:
            antecedent = self.parse_proposition(data["_antecedent"])
            consequent = self.parse_proposition(data["_consequent"])
            return ImplicationProposition(antecedent, consequent)
        if data["_type"] == Proposition.NUMBER_OF_ALTERNATIVES:
            individual = self.parse_individual(data["_content"])
            return NumberOfAlternativesProposition(individual)
        if data["_type"] == Proposition.QUESTION_STATUS:
            question = self.parse_question(data["_content"])
            status = self.parse_action_status(data["_status"])
            return QuestionStatusProposition(question, status)
        if data["_type"] == Proposition.MUTE:
            return MuteProposition()
        if data["_type"] == Proposition.UNMUTE:
            return UnmuteProposition()
        if data["_type"] == Proposition.QUIT:
            return QuitProposition()
        raise JSONParseFailure(f"Cannot parse {data} as proposition.")

    def _create_predicate_proposition(self, data):
        if data["predicate"]["sort"]["_name"] == "boolean":
            individual = None
        elif data["predicate"]["sort"]["_name"] == "string" and data["individual"] is None:
            individual = None
        else:
            individual = self.parse_individual(data["individual"])
        predicate = self.parse_predicate(data["predicate"])
        polarity = data.get("_polarity")
        return PredicateProposition(predicate, individual, polarity)

    def parse_individual(self, data):
        if data is None:
            return None
        sort_data = data["sort"]
        value_data = data["value"]
        value, sort = create_individual(value_data, sort_data)
        polarity = data["polarity"]
        ontology_name = data["_ontology_name"]
        if polarity == Polarity.POS:
            return Individual(ontology_name, value, sort)
        return NegativeIndividual(ontology_name, value, sort)

    def parse_goal(self, data):
        if data is None:
            return None
        if data.get("_goal_type") == Goal.PERFORM_GOAL:
            target = data.get("_target")
            action = self.parse_action(data["_content"])
            goal = PerformGoal(action, target=target)
            return goal
        if data.get("_goal_type") == Goal.HANDLE_GOAL:
            event = data.get("_device_event")
            ontology_name = data.get("_ontology_name")
            return HandleGoal(ontology_name, event)
        if data.get("_goal_type") == Goal.RESOLVE_GOAL:
            target = data.get("_target")
            question = self.parse_question(data["_content"])
            return ResolveGoal(question, target=target)
        raise JSONParseFailure(f"could not parse {data} as goal.")

    def parse_acc_icm(self, data, kwargs):
        if data.get("content"):
            content = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.ACC, content, content_speaker=content_speaker, **kwargs)

        return move.ICMMove(move.ICMMove.ACC, **kwargs)

    def parse_accommodate_icm(self, data, kwargs):
        if data.get("content"):
            content = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.ACCOMMODATE, content, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.ACCOMMODATE, **kwargs)

    def parse_resume_icm(self, data, kwargs):
        goal = self.parse_goal(data["content"])
        content_speaker = data.get("content_speaker")
        return move.ICMMoveWithContent(move.ICMMove.RESUME, goal, content_speaker=content_speaker, **kwargs)

    def parse_reraise_icm(self, data, kwargs):
        if data.get("content"):
            action = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.RERAISE, action, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.RERAISE, **kwargs)

    def parse_per_icm(self, data, kwargs):
        string = data.get("content", None)
        if string:
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.PER, string, content_speaker=content_speaker, **kwargs)
        return move.ICMMove(move.ICMMove.PER, **kwargs)

    def parse_und_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(move.ICMMove.UND, content, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.UND, **kwargs)

    def parse_sem_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse_proposition(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(move.ICMMove.SEM, content, content_speaker=content_speaker, **kwargs)
        content_speaker = data.get("content_speaker")
        return move.ICMMove(move.ICMMove.SEM, **kwargs)

    def parse_loadplan_icm(self, data, kwargs):
        return move.ICMMove(move.ICMMove.LOADPLAN, **kwargs)

    def parse_cardinal_sequencing_icm(self, data, kwargs):
        content_speaker = data.get("content_speaker")
        return move.ICMMoveWithContent(
            move.ICMMove.CARDINAL_SEQUENCING, int(data["content"]), content_speaker=content_speaker, **kwargs
        )

    def parse_report_inference_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(
                move.ICMMove.REPORT_INFERENCE, content, content_speaker=content_speaker, **kwargs
            )

    def parse_greet_move(self, kwargs):
        return move.GreetMove(**kwargs)

    def parse_insult_move(self, kwargs):
        return move.InsultMove(**kwargs)

    def parse_insult_response_move(self, kwargs):
        return move.InsultResponseMove(**kwargs)

    def parse_mute_move(self, kwargs):
        return move.MuteMove(**kwargs)

    def parse_unmute_move(self, kwargs):
        return move.UnmuteMove(**kwargs)

    def parse_quit_move(self, kwargs):
        return move.QuitMove(**kwargs)

    def parse_thank_you_move(self, kwargs):
        return move.ThankYouMove(**kwargs)

    def parse_thank_you_response_move(self, kwargs):
        return move.ThankYouResponseMove(**kwargs)

    def parse_prereport_move(self, data):
        service_action = data["service_action"]
        arguments = data["arguments"]
        argument_list = [self.parse_proposition(argument) for argument in arguments]
        ontology_name = data["ontology_name"]
        return move.PrereportMove(ontology_name, service_action, argument_list)

    def parse_report_move(self, data, kwargs):
        content = self.parse(data["content"])
        return move.ReportMove(content, **kwargs)

    def parse_ask_move(self, data, kwargs):
        question = self.parse_question(data["content"])
        return move.AskMove(question, **kwargs)

    def parse_service_action_outcome(self, json):
        if json["service_action_outcome"]:
            return SuccessfulServiceAction()
        return FailedServiceAction(json["failure_reason"])

    def parse_action_status(self, json):
        if json["action_status"] == ActionStatus.DONE:
            return Done()
        if json["action_status"] == ActionStatus.ABORTED:
            return Aborted(json["reason"])


class CheckingJSONParser(NonCheckingJSONParser):
    def __init__(self, ddd_name, ontology, domain_name):
        self.ontology = ontology
        self.ontology_name = ontology.get_name()
        self.domain_name = domain_name

    def parse_action(self, data):
        action_name = data["value"]
        return self.ontology.create_action(action_name)

    def _create_predicate_proposition(self, data):
        if data["predicate"]["sort"]["_name"] == "boolean":
            individual = None
        elif data["predicate"]["sort"]["_name"] == "string" and data["individual"] is None:
            individual = None
        else:
            individual = self.parse_individual(data["individual"])
        predicate = self.parse_predicate(data["predicate"])
        polarity = data.get("_polarity")
        return self.ontology.create_predicate_proposition(predicate, individual, polarity)

    def parse_predicate(self, data):
        predicate_name = data["name"]
        return self.ontology.get_predicate(predicate_name)

    def parse_findout_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_FINDOUT])
        allow_answer_from_pcom = data.get("allow_answer_from_pcom", False)
        return plan_item.Findout(self.domain_name, question, allow_answer_from_pcom)

    def parse_raise_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_RAISE])
        return plan_item.Raise(self.domain_name, question)

    def parse_individual(self, data):
        def extract(string, prefix, suffix):
            if value.startswith(prefix) and value.endswith(suffix):
                return value[len(prefix):len(value) - len(suffix)]

        def extract_datetime_value():
            return extract(value, "datetime(", ")")

        def extract_person_name_value():
            return extract(value, "person_name(", ")")

        if data is None:
            return None
        sort = self.ontology.get_sort(data["sort"]["_name"])
        value = data["value"]
        polarity = data["polarity"]
        if sort.is_datetime_sort():
            value = DateTime(extract_datetime_value())
        if sort.is_real_sort():
            value = float(value)
        if sort.is_person_name_sort():
            value = PersonName(extract_person_name_value())
        if sort.is_integer_sort():
            value = int(value)
        if polarity == Polarity.POS:
            return self.ontology.create_individual(value, sort)
        return self.ontology.create_negative_individual(value)

    def _create_lambda_abstracted_predicate_proposition(self, content):
        predicate = self.parse_predicate(content["predicate"])
        return self.ontology.create_lambda_abstracted_predicate_proposition(predicate)


class JSONTISParser():
    def __init__(self):
        self._parser = NonCheckingJSONParser()

    def parse(self, input):
        if "openqueue" in input:
            return self._parser.parse_openqueue(input["openqueue"])
        if "set" in input:
            return self._parser.parse_set(input["set"])
        if "stack" in input:
            return self.parse_stack(input)
        else:
            raise JSONParseFailure(f"cannot parse {input}.")

    def parse_stack_set(self, input):
        return stack.StackSet([self._parser.parse(item) for item in reversed(input["stackset"])])

    def parse_qud(self, input):
        contents = input["qud"]
        questions = self.parse_stack_set(contents)
        turn_information = {
            self._parser.parse(item["key"]): int(item["value"])
            for item in contents["turn_information"]
        }
        return stack.QUD(questions, turn_information)

    def parse_stack(self, input):
        return stack.Stack([self._parser.parse(item) for item in reversed(input["stack"])])

    def parse_entities(self, input):
        return Set([Entity.create_from_json(entity_as_dict) for entity_as_dict in input["set"]])

    def parse_iterators(self, input):
        return self._deserialize_iterators(input)

    def _deserialize_iterators(self, query_list):
        def try_parsing_as_list(list_or_single):
            try:
                return [self._parser.parse(value_as_json) for value_as_json in list_or_single]
            except Exception:
                return []

        def parse_as_single(single):
            return self._parser.parse(single)

        def deserialize_query(query_or_iterator):
            key_as_json = query_or_iterator["key"]
            value_as_json = query_or_iterator["value"]

            if value_as_json == []:
                value = []
            else:
                value = try_parsing_as_list(value_as_json)
                if not value:
                    value = parse_as_single(value_as_json)

            try:
                key = self._parser.parse(key_as_json)
            except AttributeError:
                key = key_as_json

            return key, value

        queries = {}
        for query_as_json in query_list:
            k, v = deserialize_query(query_as_json)
            queries[k] = v
        return queries

    def parse_last_hint_for_question(self, input):
        return {self._parser.parse(kv_dict["key"]): int(kv_dict["value"]) for kv_dict in input}

    def parse_plan(self, input):
        return JSONDomainParser().parse_plan(input)

    def parse_alts(self, input):
        return [alt for alt in input]



import logging
from tala.ddd.ddd_component_manager import DDDComponentManager
from tala.ddd.loading.component_set_loader import ComponentSetLoader
from tala.log.formats import TIS_LOGGING_COMPACT, TIS_LOGGING_NONE, TIS_LOGGING_AUTO
from tala.config import BackendConfig
from tala.utils.as_json import AsJSONMixin


class UnknownActiveDddException(Exception):
    pass


def create_odb_as_json(
    config, overridden_ddd_config_paths, should_greet, logged_tis_format, tis_update_format, log_to_stdout, log_level
):
    odb_as_json = {}

    raw_config = BackendConfig(config).read()

    odb_as_json["overridden_ddd_config_paths"] = overridden_ddd_config_paths
    odb_as_json["path"] = config or BackendConfig.default_name()
    odb_as_json["repeat_questions"] = raw_config["repeat_questions"]
    odb_as_json["ddd_names"] = raw_config["ddds"]
    odb_as_json["active_ddd_name"] = raw_config["active_ddd"]
    odb_as_json["rerank_amount"] = raw_config["rerank_amount"]
    odb_as_json["response_timeout"] = raw_config["response_timeout"]
    odb_as_json["confidence_thresholds"] = raw_config["confidence_thresholds"]
    odb_as_json["confidence_prediction_thresholds"] = raw_config["confidence_prediction_thresholds"]
    odb_as_json["short_timeout"] = raw_config["short_timeout"]
    odb_as_json["medium_timeout"] = raw_config["medium_timeout"]
    odb_as_json["long_timeout"] = raw_config["long_timeout"]
    odb_as_json["should_greet"] = should_greet
    odb_as_json["logged_tis_format"] = logged_tis_format
    odb_as_json["logged_tis_update_format"] = tis_update_format
    odb_as_json["log_to_stdout"] = log_to_stdout
    odb_as_json["log_level"] = log_level

    ddds = load_ddds(odb_as_json["ddd_names"], overridden_ddd_config_paths, raw_config["rerank_amount"])

    odb_as_json["ddds"] = [ddd.as_json() for ddd in ddds]

    _check_integrity(odb_as_json)

    return odb_as_json


def _check_integrity(odb_as_json):
    def check_active_ddd():
        if odb_as_json["active_ddd_name"] is None:
            raise UnknownActiveDddException(
                f"Expected active_ddd to be defined in backend config ({odb_as_json['ddd_names']}), "
                "but active_ddd was undefined."
            )

        if odb_as_json["active_ddd_name"] not in odb_as_json["ddd_names"]:
            raise UnknownActiveDddException(
                "Expected active_ddd as one of %s, but got %r." %
                (odb_as_json["active_ddd_name"], odb_as_json["ddd_names"])
            )

    check_active_ddd()


def load_ddds(ddd_names, overridden_config_paths, rerank_amount):
    ddd_component_manager = DDDComponentManager()
    component_set_loader = ComponentSetLoader(ddd_component_manager, overridden_config_paths)
    ddds = component_set_loader.ddds_as_list(ddd_names, rerank_amount=rerank_amount)
    return ddds


class OrchestratedDomainBundle(AsJSONMixin):
    def __init__(self, argument_dict):
        self._overridden_ddd_config_paths = argument_dict["overridden_ddd_config_paths"]
        self._path = argument_dict["path"]
        self._repeat_questions = argument_dict["repeat_questions"]
        self._ddd_names = argument_dict["ddd_names"]
        self._rerank_amount = float(argument_dict["rerank_amount"])
        self._response_timeout = argument_dict["response_timeout"]
        self._confidence_thresholds = argument_dict["confidence_thresholds"]
        self._confidence_prediction_thresholds = argument_dict["confidence_prediction_thresholds"]
        self._short_timeout = float(argument_dict["short_timeout"])
        self._medium_timeout = float(argument_dict["medium_timeout"])
        self._long_timeout = float(argument_dict["long_timeout"])
        self._should_greet = argument_dict["should_greet"]
        self._logged_tis_format = argument_dict["logged_tis_format"]
        self._logged_tis_update_format = argument_dict["logged_tis_update_format"]
        self._log_to_stdout = argument_dict["log_to_stdout"]
        self._log_level = argument_dict["log_level"]

        self.ddds_as_json = argument_dict["ddds"]

        self._active_ddd_name = argument_dict["active_ddd_name"]

        for key in self._confidence_thresholds:
            self._confidence_thresholds[key] = float(self._confidence_thresholds[key])
        for key in self._confidence_prediction_thresholds:
            self._confidence_prediction_thresholds[key] = float(self._confidence_prediction_thresholds[key])

        if self._logged_tis_format == TIS_LOGGING_AUTO:
            log_level = logging.getLevelName(self._log_level)
            self._logged_tis_format = self._generate_tis_format_from_log_level(log_level)

        self.logger = None

    def as_dict(self):
        dict_ = {}

        dict_["overridden_ddd_config_paths"] = self._overridden_ddd_config_paths
        dict_["path"] = self._path
        dict_["repeat_questions"] = self._repeat_questions
        dict_["ddd_names"] = self._ddd_names
        dict_["active_ddd_name"] = self.active_ddd
        dict_["rerank_amount"] = self._rerank_amount
        dict_["response_timeout"] = self._response_timeout
        dict_["confidence_thresholds"] = self._confidence_thresholds
        dict_["confidence_prediction_thresholds"] = self._confidence_prediction_thresholds
        dict_["short_timeout"] = self._short_timeout
        dict_["medium_timeout"] = self._medium_timeout
        dict_["long_timeout"] = self._long_timeout
        dict_["should_greet"] = self._should_greet
        dict_["logged_tis_format"] = self._logged_tis_format
        dict_["logged_tis_update_format"] = self._logged_tis_update_format
        dict_["log_to_stdout"] = self._log_to_stdout
        dict_["log_level"] = self._log_level
        dict_["ddds"] = self.ddds_as_json

        return dict_

    def __repr__(self):
        return str(self.as_dict())

    @staticmethod
    def _generate_tis_format_from_log_level(log_level):
        if log_level == logging.DEBUG:
            return TIS_LOGGING_COMPACT
        return TIS_LOGGING_NONE

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    @property
    def active_ddd(self):
        return self._active_ddd_name

    @active_ddd.setter
    def active_ddd(self, active_ddd_name):
        self._active_ddd_name = active_ddd_name

    @property
    def should_greet(self):
        return self._should_greet

    @property
    def logged_tis_format(self):
        return self._logged_tis_format

    @property
    def logged_tis_update_format(self):
        return self._logged_tis_update_format

    @property
    def log_to_stdout(self):
        return self._log_to_stdout

    @property
    def log_level(self):
        return self._log_level

    @property
    def confidence_thresholds(self):
        return self._confidence_thresholds

    @property
    def confidence_prediction_thresholds(self):
        return self._confidence_prediction_thresholds

    @property
    def overridden_ddd_config_paths(self):
        return self._overridden_ddd_config_paths

    @property
    def ddds(self):
        return self._ddds

    @ddds.setter
    def ddds(self, ddds):
        self._ddds = ddds

    @property
    def repeat_questions(self):
        return self._repeat_questions

    @property
    def ddd_names(self):
        return self._ddd_names

    @property
    def rerank_amount(self):
        return self._rerank_amount

    @property
    def response_timeout(self):
        return self._response_timeout

    @property
    def path(self):
        return self._path

    @property
    def short_timeout(self):
        return self._short_timeout

    @property
    def medium_timeout(self):
        return self._medium_timeout

    @property
    def long_timeout(self):
        return self._long_timeout



from tala.model.domain import DomainError
from tala.model.individual import Individual
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition, \
    LambdaAbstractedGoalProposition
from tala.model.ontology import OntologyError
from tala.model.polarity import Polarity
from tala.model.proposition import Proposition, PredicateProposition, GoalProposition, PropositionSet, \
    PreconfirmationProposition, UnderstandingProposition, KnowledgePreconditionProposition
from tala.model.question import KnowledgePreconditionQuestion, ConsequentQuestion


class UnknownObjectClassException(Exception):
    pass


class UnexpectedObjectClassException(Exception):
    pass


class InvalidSemanticObjectException(Exception):
    pass


class SemanticLogic(object):
    def __init__(self, ddd_component_manager):
        self._ddd_component_manager = ddd_component_manager

    def is_relevant(self, semantic_object, other_semantic_object):
        logic_object = self.from_semantic_object(semantic_object)
        return logic_object.relevant(other_semantic_object)

    def are_combinable(self, semantic_object, other_semantic_object):
        logic_object = self.from_semantic_object(semantic_object)
        return logic_object.is_combinable_with(other_semantic_object)

    def combine(self, semantic_object, other_semantic_object):
        logic_object = self.from_semantic_object(semantic_object)
        return logic_object.combine_with(other_semantic_object)

    def resolves(self, semantic_object, other_semantic_object):
        logic_object = self.from_semantic_object(semantic_object)
        return logic_object.is_resolved_by(other_semantic_object)

    def from_semantic_object(self, semantic_object):
        if semantic_object.is_lambda_abstracted_predicate_proposition():
            ontology = self._ontology_of_semantic_object(semantic_object)
            return LambdaAbstractedPredicatePropositionLogic(ontology, semantic_object)
        if semantic_object.is_lambda_abstracted_goal_proposition():
            return LambdaAbstractedGoalPropositionLogic(semantic_object)
        if semantic_object.is_question():
            if semantic_object.is_knowledge_precondition_question():
                return KnowledgePreconditionQuestionLogic(semantic_object, self)
            if semantic_object.is_consequent_question():
                ontology = self._ontology_of_semantic_object(semantic_object)
                return ConsequentQuestionLogic(ontology, semantic_object)
            return self.from_semantic_object(semantic_object.get_content())
        if semantic_object.is_proposition():
            if semantic_object.is_predicate_proposition():
                ontology = self._ontology_of_semantic_object(semantic_object)
                return PredicatePropositionLogic(ontology, semantic_object)
            if semantic_object.is_goal_proposition():
                return GoalPropositionLogic(semantic_object)
            if semantic_object.is_proposition_set():
                return PropositionSetLogic(self._ddd_component_manager, semantic_object, self)
            if semantic_object.is_preconfirmation_proposition():
                return PreconfirmationPropositionLogic(semantic_object)
            if semantic_object.is_understanding_proposition():
                return UnderstandingPropositionLogic(semantic_object)
        raise UnknownObjectClassException("Expected semantic object of known subclass but got %r" % semantic_object)

    def _ontology_of_semantic_object(self, semantic_object):
        if semantic_object.is_ontology_specific():
            ontology = self._ddd_component_manager.get_ontology(semantic_object.ontology_name)
            return ontology
        raise InvalidSemanticObjectException("Semantic object %s does not belong to an ontology" % semantic_object)


class SemanticObjectSpecificLogic(object):
    TARGET_CLASS = None

    def __init__(self, semantic_object):
        if not isinstance(semantic_object, self.TARGET_CLASS):
            raise UnexpectedObjectClassException(
                "Expected object of class %s but got %s" % (self.TARGET_CLASS, semantic_object)
            )
        self._semantic_object = semantic_object

    def relevant(self, answer):
        raise NotImplementedError("This method needs to be implemented in a subclass.")

    def is_combinable_with(self, answer):
        raise NotImplementedError("This method needs to be implemented in a subclass.")

    def combine_with(self, answer):
        raise NotImplementedError("This method needs to be implemented in a subclass.")

    def is_resolved_by(self, answer):
        raise NotImplementedError("This method needs to be implemented in a subclass.")


class SemanticObjectAndOntologySpecificLogic(SemanticObjectSpecificLogic):
    def __init__(self, ontology, semantic_object):
        SemanticObjectSpecificLogic.__init__(self, semantic_object)
        self._ontology = ontology


class LambdaAbstractedPredicatePropositionLogic(SemanticObjectAndOntologySpecificLogic):
    TARGET_CLASS = LambdaAbstractedPredicateProposition

    def relevant(self, answer):
        is_relevant_individual = answer.is_individual() \
            and self._relevant_feature_for_individual(answer)
        is_relevant_predicate_proposition = answer.is_predicate_proposition() \
            and answer.getPredicate().is_feature_of(self._semantic_object.predicate)
        return self.is_combinable_with(answer) or is_relevant_individual or is_relevant_predicate_proposition

    def _relevant_feature_for_individual(self, individual):
        for predicate in list(self._ontology.get_predicates().values()):
            if predicate.is_feature_of(self._semantic_object.predicate)\
               and predicate.getSort() == individual.getSort():
                return predicate

    def is_combinable_with(self, answer):
        try:
            self.combine_with(answer)
            return True
        except (AttributeError, DomainError, OntologyError):
            return False

    def combine_with(self, answer):
        if (
            answer.is_proposition() and (
                answer.getPredicate() == self._semantic_object.getPredicate()
                or answer.getPredicate().is_feature_of(self._semantic_object.getPredicate())
            )
        ):
            return answer
        elif answer.is_individual():
            if answer.is_positive():
                polarity = Polarity.POS
                individual = answer
            else:
                polarity = Polarity.NEG
                individual = Individual(answer.ontology_name, answer.getValue(), answer.getSort())

            feature_predicate = self._relevant_feature_for_individual(individual)
            if feature_predicate:
                output_predicate = feature_predicate
            else:
                output_predicate = self._semantic_object.predicate

            return PredicateProposition(output_predicate, individual, polarity)
        else:
            raise DomainError("cannot combine '" + str(self) + "' with '" + str(answer) + "'")

    def is_resolved_by(self, answer):
        try:
            return (answer.getPredicate() == self._semantic_object.getPredicate() and answer.is_positive())
        except AttributeError:
            pass
        try:
            return answer.getSort() == self._semantic_object.getSort() and answer.is_positive()
        except AttributeError:
            return False


class LambdaAbstractedGoalPropositionLogic(SemanticObjectSpecificLogic):
    TARGET_CLASS = LambdaAbstractedGoalProposition

    def relevant(self, answer):
        try:
            return answer.is_goal_proposition()
        except AttributeError:
            return False

    def combine_with(self, answer):
        if answer.is_goal_proposition():
            return answer

    def is_combinable_with(self, answer):
        try:
            return answer.is_goal_proposition()
        except AttributeError:
            return False

    def is_resolved_by(self, answer):
        try:
            if answer.is_goal_proposition():
                return True
        except AttributeError:
            pass
        return False


class PropositionLogic(SemanticObjectSpecificLogic):
    TARGET_CLASS = Proposition

    def relevant(self, answer):
        return self.is_combinable_with(answer)

    def is_resolved_by(self, answer):
        return self.is_combinable_with(answer)

    def is_combinable_with(self, answer):
        raise NotImplementedError("Method is_combinable_with() needs to be implemented in a subclass.")

    def combine_with(self, answer):
        raise NotImplementedError("Method combine_with() needs to be implemented in a subclass.")


class OntologySpecificPropositionLogic(PropositionLogic, SemanticObjectAndOntologySpecificLogic):
    def __init__(self, ontology, semantic_object):
        SemanticObjectAndOntologySpecificLogic.__init__(self, ontology, semantic_object)


class PredicatePropositionLogic(PropositionLogic, SemanticObjectAndOntologySpecificLogic):
    TARGET_CLASS = PredicateProposition

    def relevant(self, answer):
        return (
            self.is_combinable_with(answer) or self._has_feature_combinable_with(answer)
            or self._semantic_object._is_feature(answer)
        )

    def _has_feature_combinable_with(self, answer):
        return (
            answer.is_individual() and any([
                predicate for predicate in list(self._ontology.get_predicates().values())
                if predicate.is_feature_of(self._semantic_object.predicate) and predicate.getSort() == answer.getSort()
            ])
        )

    def is_combinable_with(self, answer):
        if answer.is_yes() or answer.is_no():
            return True
        elif answer.is_predicate_proposition() and self._semantic_object.getPredicate() == answer.getPredicate() and \
                self._semantic_object.getArgument() == answer.getArgument():
            return True
        elif answer.is_individual() and self._is_combinable_with_individual(answer):
            return True
        else:
            return False

    def combine_with(self, answer):
        if answer.is_yes():
            return self._semantic_object
        elif answer.is_no():
            return self._semantic_object.negate()
        elif answer.is_proposition():
            return self._combine_with_proposition(answer)
        elif answer.is_individual():
            return self._combine_with_individual(answer)

    def _is_combinable_with_individual(self, individual):
        if self._semantic_object.getPredicate().getSort() == individual.getSort() \
           and self._semantic_object.getArgument() is not None \
           and self._semantic_object.getArgument().getValue() == individual.getValue():
            return True
        else:
            return False

    def _combine_with_proposition(self, proposition):
        if self._semantic_object.getPredicate() == proposition.getPredicate() and \
                self._semantic_object.getArgument() == proposition.getArgument():
            return proposition

    def _combine_with_individual(self, individual):
        argument = self._semantic_object.getArgument()
        if argument == individual:
            return self._semantic_object
        if argument.getValue() == individual.getValue() \
                and argument.is_positive() != individual.is_positive():
            return self._semantic_object.negate()


class GoalPropositionLogic(SemanticObjectSpecificLogic):
    TARGET_CLASS = GoalProposition

    def is_resolved_by(self, answer):
        return self.is_combinable_with(answer)

    def relevant(self, answer):
        return self.is_combinable_with(answer)

    def is_combinable_with(self, answer):
        if answer.is_yes() or answer.is_no():
            return True
        elif answer.is_proposition() and answer.is_goal_proposition() and answer.get_goal(
        ) == self._semantic_object.get_goal():
            return True
        else:
            return False

    def combine_with(self, answer):
        if answer.is_yes():
            return self._semantic_object
        elif answer.is_no():
            return self._semantic_object.negate()
        elif answer.is_goal_proposition() and answer.get_goal() == self._semantic_object.get_goal():
            return answer
        else:
            return False


class PropositionSetLogic(SemanticObjectSpecificLogic):
    TARGET_CLASS = PropositionSet

    def __init__(self, ddd_component_manager, semantic_object, semantic_logic):
        SemanticObjectSpecificLogic.__init__(self, semantic_object)
        self._ddd_component_manager = ddd_component_manager
        self._semantic_logic = semantic_logic

    def is_combinable_with(self, answer):
        if self._semantic_object.is_single_alt():
            return self._semantic_logic.are_combinable(self._semantic_object.get_single_alt(), answer)
        elif answer.is_no():
            return True
        elif not answer.is_yes():
            for proposition in self._semantic_object.get_propositions():
                if self._semantic_logic.are_combinable(proposition, answer):
                    return True
        return False

    def is_resolved_by(self, answer):
        if self._semantic_object.is_goal_alts() and self._is_goal_proposition(answer):
            if self._some_goal_in_alts_dominates_goal(answer.get_goal()):
                return True
        if self._semantic_object.is_multi_alt():
            if answer.is_no():
                return True
            elif self._semantic_object.negate() == answer:
                return True
        if self._semantic_object.is_single_alt():
            return self._semantic_logic.resolves(self._semantic_object.get_single_alt(), answer)
        if (answer.is_proposition() or answer.is_individual()) \
           and answer.is_positive():
            return self.combine_with(answer) is not None
        return False

    def _is_goal_proposition(self, proposition):
        try:
            return proposition.is_goal_proposition()
        except AttributeError:
            pass
        return False

    def _some_goal_in_alts_dominates_goal(self, subgoal):
        for proposition in self._semantic_object._propositions:
            if proposition.is_goal_proposition():
                supergoal = proposition.get_goal()
                if self._has_domain() and self._domain.dominates(supergoal, subgoal):
                    return True
        return False

    def _has_domain(self):
        return self._semantic_object.is_ontology_specific()

    @property
    def _domain(self):
        ddd_components = self._ddd_component_manager.get_ddd_specific_components_of_ontology_name(
            self._semantic_object.ontology_name
        )
        return ddd_components.domain

    def combine_with(self, answer):
        if self._semantic_object.is_single_alt():
            return self._semantic_logic.combine(self._semantic_object.get_single_alt(), answer)
        elif answer.is_no():
            return self._semantic_object.negate()
        elif not answer.is_yes():
            for proposition in self._semantic_object.get_propositions():
                if self._semantic_logic.are_combinable(proposition, answer):
                    return self._semantic_logic.combine(proposition, answer)

    def relevant(self, answer):
        if self._semantic_object.is_multi_alt() and answer.is_no():
            return True
        elif self._semantic_object.is_single_alt():
            return self._semantic_logic.is_relevant(self._semantic_object.get_single_alt(), answer)
        elif not answer.is_yes():
            for proposition in self._semantic_object.get_propositions():
                if self._semantic_logic.is_relevant(proposition, answer):
                    return True
        return False


class PreconfirmationPropositionLogic(PropositionLogic):
    TARGET_CLASS = PreconfirmationProposition

    def is_combinable_with(self, answer):
        if answer.is_yes() or answer.is_no():
            return True
        return False

    def combine_with(self, answer):
        if answer.is_yes():
            return self._semantic_object
        elif answer.is_no():
            return self._semantic_object.negate()


class UnderstandingPropositionLogic(PropositionLogic):
    TARGET_CLASS = UnderstandingProposition

    def is_combinable_with(self, answer):
        if answer.is_yes() or answer.is_no():
            return True
        else:
            return False

    def combine_with(self, answer):
        if answer.is_yes():
            return self._semantic_object
        if answer.is_no():
            return self._semantic_object.negate()


class KnowledgePreconditionQuestionLogic(PropositionLogic):
    TARGET_CLASS = KnowledgePreconditionQuestion

    def __init__(self, semantic_object, semantic_logic):
        SemanticObjectSpecificLogic.__init__(self, semantic_object)
        self._semantic_logic = semantic_logic

    def combine_with(self, answer):
        if answer.is_yes():
            return KnowledgePreconditionProposition(self._semantic_object.get_content(), Polarity.POS)
        if answer.is_no():
            return KnowledgePreconditionProposition(self._semantic_object.get_content(), Polarity.NEG)
        try:
            return self._semantic_logic.combine(self._semantic_object.get_content(), answer)
        except (DomainError, OntologyError, AttributeError):
            pass

    def is_combinable_with(self, answer):
        if self.combine_with(answer):
            return True

    def relevant(self, answer):
        if self.combine_with(answer):
            return True


class ConsequentQuestionLogic(SemanticObjectAndOntologySpecificLogic):
    TARGET_CLASS = ConsequentQuestion

    def combine_with(self, answer):
        if answer.is_proposition() and answer.is_implication_proposition() and \
                answer.antecedent == self._semantic_object.get_content().antecedent and \
                answer.consequent.getPredicate() == self._semantic_object.get_content().consequent_predicate:
            return answer

    def is_combinable_with(self, answer):
        if self.combine_with(answer):
            return True

    def relevant(self, answer):
        if self.combine_with(answer):
            return True

    def is_resolved_by(self, answer):
        if self.combine_with(answer):
            return True



from mock import Mock
import unittest

from tala.utils.observable import Observable, Observer


class ObservableTests(unittest.TestCase):
    def test_add_observer_and_update_it(self):
        self.given_mocked_observer()
        self.given_observable()
        self.given_mocked_observer_added_to_observable()
        self.when_setting_value_in_observable("value")
        self.then_observer_was_updated_with("value")

    def given_mocked_observer(self):
        self._mock_observer = Mock(spec=Observer)

    def given_observable(self):
        self._observable = Observable()

    def given_mocked_observer_added_to_observable(self):
        self._observable.add_observer(self._mock_observer)

    def when_setting_value_in_observable(self, value):
        self._observable.set(value)

    def then_observer_was_updated_with(self, value):
        self._mock_observer.update.assert_called_once_with(value)

    def test_removed_observer_is_not_updated(self):
        self.given_mocked_observer()
        self.given_observable()
        self.given_mocked_observer_added_to_observable()
        self.given_mocked_observer_removed_from_observable()
        self.when_setting_value_in_observable("value")
        self.then_observer_was_not_updated()

    def given_mocked_observer_removed_from_observable(self):
        self._observable.remove_observer(self._mock_observer)

    def then_observer_was_not_updated(self):
        self._mock_observer.update.assert_not_called()

    def test_remove_observer_on_its_update(self):
        self.given_mocked_observer()
        self.given_observable()
        self.given_mocked_observer_added_to_observable()
        self.given_mocked_observer_removes_itself_from_observable_when_updated()
        self.when_setting_value_in_observable("value")
        self.then_there_was_no_exception()

    def given_mocked_observer_removes_itself_from_observable_when_updated(self):
        def remove_from_observable(value):
            self._observable.remove_observer(self._mock_observer)
        self._mock_observer.update.side_effect = remove_from_observable

    def then_there_was_no_exception(self):
        pass



import pytest

from tala.ddd.ddd_component_manager import DDDComponentManager
from tala.ddd.loading.component_set_loader import ComponentSetLoader
from tala.testing import utils as test_utils


@pytest.fixture(scope="class")
def ddd_component_manager(request):
    ddd_component_manager = DDDComponentManager()
    request.cls.ddd_component_manager = ddd_component_manager


@pytest.fixture(scope="class")
def loaded_mockup_travel(request):
    component_set_loader = ComponentSetLoader(request.cls.ddd_component_manager)
    test_utils.load_mockup_travel(component_set_loader)



from pathlib import Path


def ddd_files(path):
    patterns = ["*.xml", "*.txt", "*.py", "*.pgf", "*.json"]
    for pattern in patterns:
        for file in Path(path).rglob(pattern):
            yield file



from tala.utils.as_json import AsJSONMixin


class TISNode(AsJSONMixin):
    def pretty_string_new(self, path):
        output = ""
        for attr in sorted(self.__dict__.keys()):
            value = self.__dict__[attr]
            path.append(str(attr))
            output += self.value_string_new(path, value)
            path.pop()
        return output

    def value_string_new(self, path, value):
        try:
            return value.pretty_string_new(path)
        except AttributeError:
            string_value = str(value)
            if len(string_value) > 0 and string_value[0] != "_":
                return "\n" + ".".join(path) + ": " + string_value
            return ""

    def pretty_string(self, indentation_level):
        output = ""
        for attr in sorted(self.__dict__.keys()):
            value = self.__dict__[attr]
            output += "\n" + self.indent(indentation_level + 1)
            output += attr + ": " + self.value_string(value, indentation_level + 1)
        return output

    def value_string(self, value, indentation_level):
        try:
            return value.pretty_string(indentation_level)
        except AttributeError:
            return str(value)

    def indent(self, indentation_level):
        return "\t" * indentation_level



from tala.ddd.json_parser import JSONDDDParser
from tala.ddd.parser import Parser
from tala.ddd.services.parameters.retriever import ParameterRetriever
from tala.ddd.ddd_specific_components import DDDSpecificComponents
from tala.ddd.domain_manager import DomainManager
from tala.model.semantic_logic import SemanticLogic


class DDDSpecificComponentsAlreadyExistsException(Exception):
    pass


class UnexpectedDomainException(Exception):
    pass


class UnexpectedOntologyException(Exception):
    pass


class UnexpectedDDDException(Exception):
    pass


class DDDComponentManager(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.ddd_names = []
        self.ddds_as_json = []
        self._components_of_ddds = {}
        self.ontologies = {}
        self.domains = {}
        self.domain_manager = DomainManager(self)
        self._ddds_of_domains = {}
        self._ddds_of_ontologies = {}
        self._semantic_logic = SemanticLogic(self)

    @property
    def semantic_logic(self):
        return self._semantic_logic

    def add(self, ddd_specific_components):
        if ddd_specific_components.name in self._components_of_ddds:
            raise DDDSpecificComponentsAlreadyExistsException(
                "Components for DDD '%s' already registered" % ddd_specific_components.name
            )
        self.add_ontology(ddd_specific_components.ontology)
        self._ddds_of_ontologies[ddd_specific_components.ontology] = ddd_specific_components
        self.add_domain(ddd_specific_components.domain)
        self._ddds_of_domains[ddd_specific_components.domain] = ddd_specific_components
        self._components_of_ddds[ddd_specific_components.name] = ddd_specific_components

    def ensure_ddd_components_added(self, ddd_components):
        if ddd_components.name not in self._components_of_ddds:
            self.add(ddd_components)

    def add_domain(self, domain):
        self.domains[domain.get_name()] = domain
        self.domain_manager.add(domain)

    def add_ontology(self, ontology):
        self.ontologies[ontology.get_name()] = ontology

    def get_components_for_all_ddds(self):
        return list(self._components_of_ddds.values())

    def get_ddd_specific_components(self, name):
        if name not in self._components_of_ddds and self.ddds_as_json:
            self._load_ddd(name)
        return self._components_of_ddds[name]

    def _load_ddd(self, name):
        if name not in self.ddd_names:
            raise UnexpectedDDDException(f"Expected one of the known DDDs {self.ddd_names}, but got '{name}'")
        if not self._is_loaded(name):
            ddd_as_json = self._get_ddd_as_json(name)
            self._parse_and_add(ddd_as_json)

    def load_ddd_for_ontology_name(self, name):
        for ddd_as_json in self.ddds_as_json:
            ontology_name = ddd_as_json["ontology"]["_name"]
            if ontology_name == name:
                self._parse_and_add(ddd_as_json)
                return
        raise UnexpectedDDDException(f"Expected ontology name of a known DDD ({self.ddd_names}), but got '{name}'")

    def _is_loaded(self, ddd_name):
        return ddd_name in self._components_of_ddds

    def add_ddds_as_json(self, ddd_names, ddds_as_json):
        self.ddd_names = ddd_names
        self.ddds_as_json = ddds_as_json

    def _get_ddd_as_json(self, name):
        for ddd in self.ddds_as_json:
            if name == ddd["ddd_name"]:
                return ddd

    def _parse_and_add(self, ddd_as_json):
        ddd = JSONDDDParser().parse(ddd_as_json)
        parameter_retriever = ParameterRetriever(ddd.service_interface, ddd.ontology)
        parser = Parser(ddd.name, ddd.ontology, ddd.domain.name)
        ddd_specific_components = DDDSpecificComponents(ddd, parameter_retriever, parser)
        self.add(ddd_specific_components)

    def get_domain(self, name):
        return self.domains[name]

    def get_ontology(self, name):
        if name not in self.ontologies:
            self.load_ddd_for_ontology_name(name)
        return self.ontologies[name]

    def get_ddd_specific_components_of_ontology(self, ontology):
        if ontology not in self._ddds_of_ontologies:
            raise UnexpectedOntologyException(
                "Expected to find '%s' among known ontologies %s but did not." %
                (ontology, list(self._ddds_of_ontologies.keys()))
            )
        return self._ddds_of_ontologies[ontology]

    def get_ddd_specific_components_of_ontology_name(self, ontology_name):
        if ontology_name not in self.ontologies:
            self.load_ddd_for_ontology_name(ontology_name)
        ontology = self.get_ontology(ontology_name)
        return self.get_ddd_specific_components_of_ontology(ontology)

    def reset_components_of_ddd(self, name):
        ddd_specific_components = self._components_of_ddds[name]
        ddd_specific_components.reset()



from tala.model.goal import ResolveGoal
from tala.model.speaker import Speaker


class UnknownGoalException(Exception):
    pass


class DomainAlreadyAddedException(Exception):
    pass


class DomainManager(object):
    def __init__(self, ddd_component_manager):
        self._domains_of_goals = {}
        self._domains = set()
        self._ddd_component_manager = ddd_component_manager

    @property
    def domains(self):
        return self._domains

    def add(self, domain):
        for goal in domain.goals:
            self._domains_of_goals[goal] = domain
        self._domains.add(domain)

    def has_goal(self, goal):
        return goal in self._domains_of_goals

    def get_domain_of_goal(self, goal):
        self._load_domain_if_needed(goal)
        if goal not in self._domains_of_goals:
            if goal.is_resolve_goal():
                question = goal.get_content()
                if question.is_consequent_question():
                    return self.get_domain_of_goal(
                        ResolveGoal(question.get_embedded_consequent_question(), Speaker.SYS)
                    )
            raise UnknownGoalException(
                "Goal %s not found among known goals %s" % (repr(goal), list(self._domains_of_goals.keys()))
            )
        return self._domains_of_goals[goal]

    def _load_domain_if_needed(self, goal):
        if goal.is_ontology_specific and goal not in self._domains_of_goals:
            self._ddd_component_manager.load_ddd_for_ontology_name(goal.ontology_name)

    def remove(self, domain):
        for goal in domain.goals:
            if goal not in self._domains_of_goals:
                raise UnknownGoalException(
                    f"Goal '{goal}' of domain '{domain}' not found among known "
                    f"goals '{list(self._domains_of_goals.keys())}'"
                )
            self._domains_of_goals[goal] = None
            del self._domains_of_goals[goal]
        self._domains.remove(domain)

    def get_downdate_conditions(self, goal):
        if self.has_goal(goal):
            domain = self.get_domain_of_goal(goal)
            return domain.get_downdate_conditions(goal)
        else:
            return []

    def goal_allows_accommodation_without_feedback(self, goal):
        domain = self.get_domain_of_goal(goal)
        return domain.get_goal_attribute(goal, "accommodate_without_feedback")

    def get_all_goals_in_defined_order(self):
        goals = []
        for domain in self._domains:
            domain_goals = domain.get_all_goals_in_defined_order()
            goals.extend(domain_goals)
        return goals

    def goal_is_preferred(self, goal):
        domain = self.get_domain_of_goal(goal)
        return domain.goal_is_preferred(goal)

    def get_io_status(self, goal):
        domain = self.get_domain_of_goal(goal)
        return domain.get_io_status(goal)



import unittest

from mock import Mock

from tala.model.proposition import PredicateProposition, Proposition, ResolvednessProposition
from tala.model.predicate import Predicate
from tala.model import condition
from tala.model.condition import Condition, HasValue, IsTrue, HasSharedValue
from tala.model.set import Set


class TIS:
    def __init__(self):
        self.shared = Shared()
        self.private = Private()


class Shared:
    def __init__(self):
        self.com = {}


class Private:
    def __init__(self):
        self.bel = {}


class ConditionTestCase(unittest.TestCase):
    def setUp(self):
        self.predicate_a = Mock(spec=Predicate)
        self.predicate_b = Mock(spec=Predicate)
        self.proposition_over_predicate_a = Mock(spec=PredicateProposition)
        self.proposition_over_predicate_a.get_predicate.return_value = self.predicate_a
        self.proposition_b = Mock(spec=PredicateProposition)
        self.other_proposition = Mock(spec=Proposition)
        self.resolvedness_proposition = Mock(spec=ResolvednessProposition)

    def given_condition(self, condition):
        self._condition = condition

    def given_set_of_propositions(self, propositions):
        self._propositions = Set()
        for proposition in propositions:
            self._propositions.add(proposition)

    def when_is_true_given_proposition_set_is_called(self):
        self._result = self._condition.is_true_given_proposition_set(self._propositions)

    def then_result_is(self, value):
        assert self._result is value

    def when_tested_for_equality(self, first_object, second_object):
        self._result = first_object == second_object


class TestConditions(ConditionTestCase):
    def test_get_truth_value_given_empty_facts(self):
        self.given_condition(Condition("some_type"))
        self.given_set_of_propositions([])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_get_truth_value_given_non_empty_facts(self):
        self.given_condition(Condition("some_type"))
        self.given_set_of_propositions([self.proposition_over_predicate_a])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_equality(self):
        self.when_tested_for_equality(Condition("some_type"), Condition("some_type"))
        self.then_result_is(True)

    def test_inequality_due_to_type(self):
        self.when_tested_for_equality(Condition("some_type"), Condition("other_type"))
        self.then_result_is(False)

    def test_inequality_due_to_interface(self):
        self.when_tested_for_equality(Condition("some_type"), "random object")
        self.then_result_is(False)

    def test_inequality_due_to_none(self):
        self.when_tested_for_equality(Condition("some_type"), None)
        self.then_result_is(False)


class TestHasValueCondition(ConditionTestCase):
    def test_creation_with_standard_predicate(self):
        HasValue(self.predicate_a)

    def test_get_predicate_for_standard_predicate(self):
        self._given_hasvalue_condition_created_for_predicate(self.predicate_a)
        self._when_predicate_is_collected()
        self._then_predicate_is(self.predicate_a)

    def _given_hasvalue_condition_created_for_predicate(self, predicate):
        self._has_value_condition = HasValue(predicate)

    def _when_predicate_is_collected(self):
        self._result = self._has_value_condition.predicate

    def _then_predicate_is(self, predicate):
        self.assertEqual(predicate, self._result)

    def test_string_for_standard_predicate(self):
        self._when_string_is_generated_for_has_value_with_predicate(self.predicate_a)
        self._then_string_is(f"HasValue({self.predicate_a})")

    def _when_string_is_generated_for_has_value_with_predicate(self, predicate):
        self._string = str(HasValue(predicate))

    def _then_string_is(self, string):
        self.assertEqual(string, self._string)

    def test_type(self):
        self._given_hasvalue_condition_created_for_predicate(self.predicate_a)
        self._when_type_is_collected()
        self._then_type_is(condition.HAS_VALUE)

    def _when_type_is_collected(self):
        self._type = self._has_value_condition.condition_type

    def _then_type_is(self, type):
        self.assertEqual(type, self._type)

    def test_equality(self):
        self.when_tested_for_equality(HasValue(self.predicate_a), HasValue(self.predicate_a))
        self.then_result_is(True)

    def test_inequality_due_to_predicate(self):
        self.when_tested_for_equality(HasValue(self.predicate_a), HasValue(self.predicate_b))
        self.then_result_is(False)

    def test_is_true_given_containing_proposition_with_same_predicate(self):
        self.given_condition(HasValue(self.predicate_a))
        self.given_set_of_propositions([self.proposition_over_predicate_a])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)

    def test_is_true_given_containing_proposition_with_different_predicate(self):
        self.given_condition(HasValue(self.predicate_a))
        self.given_set_of_propositions([self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_get_truth_value_tolerates_any_proposition(self):
        self.given_condition(HasValue(self.proposition_over_predicate_a))
        self.given_set_of_propositions([self.other_proposition])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_get_truth_value_given_several_propositions(self):
        self.given_condition(HasValue(self.predicate_a))
        self.given_set_of_propositions([self.proposition_over_predicate_a, self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)


class TestIsTrueConditions(ConditionTestCase):
    def test_creation_with_standard_proposition(self):
        IsTrue(self.proposition_over_predicate_a)

    def test_get_proposition_for_standard_proposition(self):
        self._given_is_shared_fact_created_for_proposition(self.proposition_over_predicate_a)
        self._when_proposition_is_collected()
        self._then_proposition_is(self.proposition_over_predicate_a)

    def _given_is_shared_fact_created_for_proposition(self, proposition):
        self._condition = IsTrue(proposition)

    def _when_proposition_is_collected(self):
        self._result = self._condition.proposition

    def _then_proposition_is(self, proposition):
        self.assertEqual(proposition, self._result)

    def test_string_for_standard_predicate(self):
        self._when_string_is_generated_for_condition_with_proposition(self.proposition_over_predicate_a)
        self._then_string_is(f"IsTrue({self.proposition_over_predicate_a})")

    def _when_string_is_generated_for_condition_with_proposition(self, proposition):
        self._string = str(IsTrue(proposition))

    def _then_string_is(self, string):
        self.assertEqual(string, self._string)

    def test_type(self):
        self._given_is_shared_fact_condition_created_for_proposition(self.proposition_over_predicate_a)
        self._when_type_is_collected()
        self._then_type_is(condition.IS_TRUE)

    def _given_is_shared_fact_condition_created_for_proposition(self, proposition):
        self._is_shared_fact_condition = IsTrue(proposition)

    def _when_type_is_collected(self):
        self._type = self._is_shared_fact_condition.condition_type

    def _then_type_is(self, type):
        self.assertEqual(type, self._type)

    def test_equality(self):
        self.when_tested_for_equality(
            IsTrue(self.proposition_over_predicate_a), IsTrue(self.proposition_over_predicate_a)
        )
        self.then_result_is(True)

    def test_inequality_due_to_proposition(self):
        self.when_tested_for_equality(IsTrue(self.proposition_over_predicate_a), IsTrue(self.proposition_b))
        self.then_result_is(False)

    def test_inequality_due_to_type(self):
        self.when_tested_for_equality(HasValue(self.predicate_a), IsTrue(self.proposition_over_predicate_a))
        self.then_result_is(False)

    def test_inequality_due_to_interface(self):
        self.when_tested_for_equality(IsTrue(self.proposition_over_predicate_a), "random_object")
        self.then_result_is(False)

    def test_is_true_given_containing_same_proposition(self):
        self.given_condition(IsTrue(self.proposition_over_predicate_a))
        self.given_set_of_propositions([self.proposition_over_predicate_a])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)

    def test_is_true_given_containing_different_proposition(self):
        self.given_condition(IsTrue(self.proposition_over_predicate_a))
        self.given_set_of_propositions([self.other_proposition])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_is_true_given_consisting_of_several_propositions(self):
        self.given_condition(IsTrue(self.proposition_over_predicate_a))
        self.given_set_of_propositions([self.proposition_over_predicate_a, self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)


class TestHasSharedValueCondition(ConditionTestCase):
    def test_creation_with_standard_predicate(self):
        HasSharedValue(self.predicate_a)

    def test_string_for_standard_predicate(self):
        self._when_string_is_generated_for_has_value_with_predicate(self.predicate_a)
        self._then_string_is(f"HasSharedValue({self.predicate_a})")

    def _when_string_is_generated_for_has_value_with_predicate(self, predicate):
        self._string = str(HasSharedValue(predicate))

    def _then_string_is(self, string):
        self.assertEqual(string, self._string)

    def test_type(self):
        self._given_has_shared_value_condition_created_for_predicate(self.predicate_a)
        self._when_type_is_collected()
        self._then_type_is(condition.HAS_SHARED_VALUE)

    def _given_has_shared_value_condition_created_for_predicate(self, predicate):
        self._has_value_condition = HasSharedValue(predicate)

    def _when_type_is_collected(self):
        self._type = self._has_value_condition.condition_type

    def _then_type_is(self, type):
        self.assertEqual(type, self._type)

    def test_is_true_given_containing_proposition_with_same_predicate(self):
        self.given_condition(HasSharedValue(self.predicate_a))
        self.given_tis_with_commitments([self.proposition_over_predicate_a])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)

    def given_tis_with_commitments(self, commitments):
        self.tis = TIS()
        self.tis.shared.com = commitments

    def when_is_true_given_proposition_set_is_called(self):
        self._result = self._condition.is_true_given_tis(self.tis)

    def test_equality(self):
        self.when_tested_for_equality(HasSharedValue(self.predicate_a), HasSharedValue(self.predicate_a))
        self.then_result_is(True)

    def test_inequality_due_to_predicate(self):
        self.when_tested_for_equality(HasSharedValue(self.predicate_a), HasSharedValue(self.predicate_b))
        self.then_result_is(False)

    def test_is_true_given_containing_proposition_with_different_predicate(self):
        self.given_condition(HasSharedValue(self.predicate_a))
        self.given_tis_with_commitments([self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_get_truth_value_tolerates_any_proposition(self):
        self.given_condition(HasSharedValue(self.proposition_over_predicate_a))
        self.given_tis_with_commitments([self.other_proposition])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def test_get_truth_value_given_several_propositions_in_com(self):
        self.given_condition(HasSharedValue(self.predicate_a))
        self.given_tis_with_commitments([self.proposition_over_predicate_a, self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(True)

    def test_get_truth_value_given_several_propositions_in_bel(self):
        self.given_condition(HasSharedValue(self.predicate_a))
        self.given_tis_with_beliefs([self.proposition_over_predicate_a, self.proposition_b])
        self.when_is_true_given_proposition_set_is_called()
        self.then_result_is(False)

    def given_tis_with_beliefs(self, beliefs):
        self.tis = TIS()
        self.tis.private.beliefs = beliefs



import unittest

from tala.utils import unicodify


class TestSelectionTests(unittest.TestCase):
    def setUp(self):
        self.unicodify = unicodify

    def test_sequence_string(self):
        self.when_sequence_string(["item one", "item two"])
        self.then_unicode_str_is("['item one', 'item two']")

    def when_sequence_string(self, sequence):
        self.unicode_str = self.unicodify._sequence_string(sequence)

    def then_unicode_str_is(self, expected_str):
        self.assertEqual(self.unicode_str, expected_str)

    def test_sequence_string_nested(self):
        self.when_sequence_string(["item one", ["sub item one", "sub item two"]])
        self.then_unicode_str_is("['item one', ['sub item one', 'sub item two']]")

    def test_dict_string(self):
        self.when_dict_string({"value one": "key one", "value two": "key two"})
        self.then_unicode_str_is("{'value one': 'key one', 'value two': 'key two'}")

    def test_dict_string_nested(self):
        self.when_dict_string({
            "value one": {
                "sub value one": "sub key two",
                "sub value two": "sub key two"
            },
            "value two": "key two"
        })
        self.then_unicode_str_is(
            "{'value one': {'sub value one': 'sub key two', 'sub value two': 'sub key two'}, 'value two': 'key two'}"
        )

    def when_dict_string(self, _dict):
        self.unicode_str = self.unicodify._dict_string(_dict)

    def test_tuple_string(self):
        self.when_tuple_string(("item one", "item two"))
        self.then_unicode_str_is("('item one', 'item two')")

    def when_tuple_string(self, _tuple):
        self.unicode_str = self.unicodify._tuple_string(_tuple)

    def test_tuple_string_nested(self):
        self.when_tuple_string(("item one", ("sub item one", "item two")))
        self.then_unicode_str_is("('item one', ('sub item one', 'item two'))")



from tala.model.proposition import PredicateProposition


class Database:
    def __init__(self, resourceclass, ontology):
        self.resource = resourceclass()
        self.ontology = ontology

    def consultDB(self, question, commitments):
        value = self.resource.consultDB(question, commitments)
        predicate = question.get_predicate()
        individual = self.ontology.create_individual(value)
        proposition = PredicateProposition(predicate, individual)
        return proposition



import unittest

import pytest

from tala.model.semantic_logic import SemanticLogic


@pytest.mark.usefixtures("ddd_component_manager", "loaded_mockup_travel")
class SemanticsTester(unittest.TestCase):
    def setUp(self):
        self._semantic_logic = SemanticLogic(self.ddd_component_manager)
        self.ddd_components = self.ddd_component_manager.get_ddd_specific_components("mockup_travel")
        self.parser = self.ddd_components.parser
        self.valid_combinations = []
        self.valid_relevances = []
        self.valid_resolves = []
        self.valid_incompatibles = []
        self.answers = set()
        self.questions = set()
        self.failures = []

    def tearDown(self):
        if len(self.failures) > 0:
            self.fail("Failures:\n" + "\n".join(self.failures))
        self.ddd_component_manager.reset_components_of_ddd(self.ddd_components.name)

    def declare_combinable_and_relevant(self, declarations):
        for declaration in declarations:
            (question_as_string, answer_as_string, result_as_string) = \
                declaration
            question = self.create_object(question_as_string)
            answer = self.create_object(answer_as_string)
            result = self.create_object(result_as_string)
            self.valid_combinations.append((question, answer, result))
            self.valid_relevances.append((question, answer))
            self.answers.add(answer)
            self.questions.add(question)

    def declare_relevant_not_resolving(self, declarations):
        for declaration in declarations:
            (question_as_string, answer_as_string) = declaration
            question = self.create_object(question_as_string)
            answer = self.create_object(answer_as_string)
            self.valid_relevances.append((question, answer))
            self.answers.add(answer)
            self.questions.add(question)

    def declare_resolving(self, declarations):
        for declaration in declarations:
            (question_as_string, answer_as_string) = declaration
            question = self.create_object(question_as_string)
            answer = self.create_object(answer_as_string)
            self.valid_resolves.append((question, answer))
            self.answers.add(answer)
            self.questions.add(question)

    def declare_answer(self, answer_as_string):
        answer = self.create_object(answer_as_string)
        self.answers.add(answer)

    def declare_incompatible(self, declarations):
        for declaration in declarations:
            (answer_1_as_string, answer_2_as_string) = declaration
            answer_1 = self.create_object(answer_1_as_string)
            answer_2 = self.create_object(answer_2_as_string)
            self.answers.add(answer_1)
            self.answers.add(answer_2)
            self.valid_incompatibles.append((answer_1, answer_2))

    def semantics_test_valid_combines(self):
        for combination in self.valid_combinations:
            (question, answer, result) = combination
            self._assert_combination(question, answer, result)

    def semantics_test_invalid_combines(self):
        for question in self.questions:
            for answer in self.answers:
                if self._is_invalid_combination(question, answer):
                    self._assert_not_combinable(question, answer)

    def semantics_test_relevance(self):
        for relevance in self.valid_relevances:
            (question, answer) = relevance
            self._assert_relevant(question, answer)

    def semantics_test_irrelevance(self):
        for question in self.questions:
            for answer in self.answers:
                if self._is_irrelevant(question, answer):
                    self._assert_not_relevant(question, answer)

    def semantics_test_resolves(self):
        for resolve in self.valid_resolves:
            (question, answer) = resolve
            self._assert_resolving(question, answer)

    def semantics_test_non_resolves(self):
        for question in self.questions:
            for answer in self.answers:
                if self._is_not_resolving(question, answer):
                    self._assert_not_resolving(question, answer)

    def semantics_test_valid_incompatibles(self):
        for (answer_1, answer_2) in self.valid_incompatibles:
            self._assert_incompatible(answer_1, answer_2)
            self._assert_incompatible(answer_2, answer_1)

    def semantics_test_invalid_incompatibles(self):
        for answer_1 in self._get_propostional_answers():
            for answer_2 in self._get_propostional_answers():
                if not (answer_1, answer_2) in self.valid_incompatibles and \
                        not (answer_2, answer_1) in self.valid_incompatibles:
                    self._assert_compatible(answer_1, answer_2)
                    self._assert_compatible(answer_2, answer_1)

    def _is_invalid_combination(self, question, answer):
        for combination in self.valid_combinations:
            (combinable_question, combinable_answer, result) = combination
            if combinable_question == question and combinable_answer == answer:
                return False
        return True

    def _is_irrelevant(self, question, answer):
        return not (question, answer) in self.valid_relevances

    def _is_not_resolving(self, question, answer):
        return not (question, answer) in self.valid_resolves

    def _assert_combination(self, question, answer, expected_result):
        if not self._semantic_logic.are_combinable(question, answer):
            self._report_failure(
                "SemanticLogic.are_combinable(%s, %s) should be True but was False" % (question, answer)
            )
        actual_result = self._semantic_logic.combine(question, answer)
        if expected_result != actual_result:
            self._report_failure(
                "SemanticLogic.combine(%s, %s) should return %s but returned %s" %
                (question, answer, expected_result, actual_result)
            )

    def _assert_not_combinable(self, question, answer):
        if self._semantic_logic.are_combinable(question, answer):
            self._report_failure(
                "SemanticLogic.are_combinable(%s, %s) should be False but was True" % (question, answer)
            )

    def _assert_resolving(self, question, answer):
        if not self._semantic_logic.resolves(question, answer):
            self._report_failure("SemanticLogic.resolves(%s, %s) should be True but was False" % (question, answer))

    def _assert_not_resolving(self, question, answer):
        if self._semantic_logic.resolves(question, answer):
            self._report_failure(
                "SemanticLogic.resolves(%s, %s) has not been explicitly declared but was True" % (question, answer)
            )

    def _assert_relevant(self, question, answer):
        if not self._semantic_logic.is_relevant(question, answer):
            self._report_failure("relevant(%s, %s) should be True but was False" % (question, answer))

    def _assert_not_relevant(self, question, answer):
        if self._semantic_logic.is_relevant(question, answer):
            self._report_failure("relevant(%s, %s) has not been explicitly declared but was True" % (question, answer))

    def _report_failure(self, failure):
        self.failures.append(failure)

    def _get_propostional_answers(self):
        return list(filter(self._is_proposition, self.answers))

    @staticmethod
    def _is_proposition(answer):
        try:
            return answer.is_proposition()
        except AttributeError:
            return False

    def _assert_incompatible(self, proposition_1, proposition_2):
        if not proposition_1.is_incompatible_with(proposition_2):
            self._report_failure("%s should be incompatible with %s" % (proposition_1, proposition_2))

    def _assert_compatible(self, proposition_1, proposition_2):
        if proposition_1.is_incompatible_with(proposition_2):
            self._report_failure("%s should be compatible with %s" % (proposition_1, proposition_2))

    def create_object(self, string):
        return self.parser.parse(string)



import pytest

from tala.config import BackendConfig, DddConfig, DeploymentsConfig

from .console_script_mixin import ConsoleScriptTestMixin


class TestConfigFileIntegration(ConsoleScriptTestMixin):
    EXPECTED_CONFIGS = {
        BackendConfig: {
            "active_ddd": "my_ddd",
            "asr": 'none',
            "confidence_prediction_thresholds": {
                "ACKNOWLEDGE": 0.8,
                "CHECK": 0.6,
                "TRUST": 1.0
            },
            "confidence_thresholds": {
                "ACKNOWLEDGE": 0.15,
                "CHECK": 0.1,
                "TRUST": 0.3
            },
            "ddds": ["my_ddd"],
            "inactive_seconds_allowed": 7200,
            'long_timeout': 5.0,
            'medium_timeout': 2.0,
            'short_timeout': 1.0,
            "repeat_questions": True,
            "rerank_amount": 0.2,
            "response_timeout": 2.5,
            "supported_languages": 'none',
            "use_recognition_profile": False,
        },
        DddConfig: {
            'device_module': None,
            'rasa_nlu': {},
            'use_rgl': False,
            'use_third_party_parser': False,
            'word_list': 'word_list.txt',
        },
        DeploymentsConfig: {
            "dev": "https://127.0.0.1:9090/interact"
        }
    }

    EXPECTED_PARAMETERISED_CONFIGS = {
        BackendConfig: {
            "active_ddd": "my_ddd",
            "asr": 'none',
            "confidence_prediction_thresholds": {
                "ACKNOWLEDGE": 0.8,
                "CHECK": 0.6,
                "TRUST": 1.0
            },
            "confidence_thresholds": {
                "ACKNOWLEDGE": 0.15,
                "CHECK": 0.1,
                "TRUST": 0.3
            },
            "ddds": ["my_ddd"],
            "inactive_seconds_allowed": 7200,
            "repeat_questions": True,
            "rerank_amount": 0.2,
            'long_timeout': 5.0,
            'medium_timeout': 2.0,
            'short_timeout': 1.0,
            "response_timeout": 2.5,
            "supported_languages": 'none',
            "use_recognition_profile": False,
        }
    }

    @pytest.mark.parametrize(
        "ConfigClass,command", [(BackendConfig, "create-backend-config my_ddd"), (DddConfig, "create-ddd-config"),
                                (DeploymentsConfig, "create-deployments-config")]
    )
    def test_create_config_without_path(self, ConfigClass, command):
        self._when_running_command(f"tala {command}")
        self._then_config_contains(ConfigClass, ConfigClass.default_name(), self.EXPECTED_CONFIGS[ConfigClass])

    @pytest.mark.parametrize("ConfigClass,command", [(BackendConfig, "create-backend-config my_ddd")])
    def test_create_parameterised_config(self, ConfigClass, command):
        self._when_running_command(f"tala {command}")
        self._then_config_contains(
            ConfigClass, ConfigClass.default_name(), self.EXPECTED_PARAMETERISED_CONFIGS[ConfigClass]
        )

    def _then_config_contains(self, ConfigClass, name, expected_config):
        actual_config = ConfigClass(name).read()
        assert expected_config == actual_config

    @pytest.mark.parametrize(
        "ConfigClass,command", [(BackendConfig, "create-backend-config my_ddd"), (DddConfig, "create-ddd-config"),
                                (DeploymentsConfig, "create-deployments-config")]
    )
    def test_create_config_with_path(self, ConfigClass, command):
        self._when_running_command(f"tala {command} --filename my_ddd.config.json")
        self._then_config_contains(ConfigClass, "my_ddd.config.json", self.EXPECTED_CONFIGS[ConfigClass])

    @pytest.mark.parametrize(
        "name,command", [
            ("backend", "create-backend-config mock_ddd"),
            ("DDD", "create-ddd-config"),
            ("deployments", "create-deployments-config"),
        ]
    )
    def test_exception_raised_if_config_file_already_exists(self, name, command):
        self._given_config_was_created_with("tala {} --filename test.config.json".format(command))
        self._when_running_command("tala {} --filename test.config.json".format(command))
        self._then_stderr_contains(
            "Expected to be able to create {} config file 'test.config.json' but it already exists.".format(name)
        )

    def _given_config_was_created_with(self, command):
        self._run_command(command)

    @pytest.mark.parametrize(
        "command", [
            "create-backend-config mock_ddd",
            "create-ddd-config",
            "create-deployments-config",
        ]
    )
    def test_config_file_not_overwritten(self, command):
        self._given_file_contains("test.config.json", "unmodified_mock_content")
        self._when_running_command("tala {} --filename test.config.json".format(command))
        self._then_file_contains("test.config.json", "unmodified_mock_content")



from tala.ddd.services.service_interface import ParameterField, ParameterNotFoundException
from tala.ddd.services.parameters.binding import SingleInstanceParameterBinding, MultiInstanceParameterBinding


class MissingRequiredParameterException(Exception):
    pass


class UnexpectedNumberOfFactsException(Exception):
    pass


class ParameterRetriever(object):
    def __init__(self, service_interface, ontology):
        super(ParameterRetriever, self).__init__()
        self._service_interface = service_interface
        self._ontology = ontology

    def preprocess_validity_arguments(self, name, facts, session):
        validator = self._service_interface.get_validator(name)
        return self.preprocess_arguments(validator.parameters, facts, session)

    def preprocess_query_arguments(self, name, facts, session):
        query = self._service_interface.get_query(name)
        return self.preprocess_arguments(query.parameters, facts, session)

    def preprocess_action_arguments(self, name, facts, session):
        action = self._service_interface.get_action(name)
        return self.preprocess_arguments(action.parameters, facts, session)

    def preprocess_arguments(self, parameters, facts, session):
        return [self.preprocess_argument(parameter, facts, session) for parameter in parameters]

    def preprocess_argument(self, parameter, facts, session):
        def single_instance_parameter_binding(parameter_facts):
            if len(parameter_facts) == 0:
                return single_instance_parameter_binding_without_fact()
            elif len(parameter_facts) == 1:
                return single_instance_parameter_binding_with_fact(parameter_facts[0])
            else:
                raise UnexpectedNumberOfFactsException(
                    f"Expected 0 or 1 facts for predicate {parameter.name} but got {len(parameter_facts)}: "
                    f"{parameter_facts}"
                )

        def single_instance_parameter_binding_without_fact():
            if not parameter.is_optional:
                raise ParameterNotFoundException("failed to get argument '%s'" % parameter.name)
            return SingleInstanceParameterBinding(parameter=parameter, value=None, proposition=None)

        def single_instance_parameter_binding_with_fact(proposition):
            value = self._get_argument_value(parameter, proposition, session)
            return SingleInstanceParameterBinding(parameter, value, proposition)

        parameter_facts = list(self.get_facts_by_predicate_name(parameter.name, facts))
        predicate = self._ontology.get_predicate(parameter.name)
        if predicate.allows_multiple_instances():
            values = [self._get_argument_value(parameter, fact, session) for fact in parameter_facts]
            return MultiInstanceParameterBinding(parameter, values, parameter_facts)
        return single_instance_parameter_binding(parameter_facts)

    def _get_argument_value(self, parameter, fact, session):
        def get_grammar_entry_for_value_of_custom_sort(value):
            try:
                entities = session["entities"]
                for entity in entities:
                    if entity["name"] == value:
                        return entity["natural_language_form"]
            except KeyError:
                return None

        individual = fact.getArgument()
        sort = individual.getSort()
        value = individual.getValue()
        if parameter.format == ParameterField.VALUE:
            return sort.value_as_basic_type(value)
        elif parameter.format == ParameterField.GRAMMAR_ENTRY:
            if sort.is_builtin():
                return None
            else:
                return get_grammar_entry_for_value_of_custom_sort(value)
        else:
            raise Exception("unknown parameter field %r" % parameter.format)

    @classmethod
    def get_facts_by_predicate_name(cls, predicate_name, facts):
        for fact in cls.positive_predicate_proposition_facts(facts):
            if fact.getPredicate().get_name() == predicate_name:
                yield fact

    @classmethod
    def predicate_names_of_facts(cls, facts):
        predicate_facts = cls.positive_predicate_proposition_facts(facts)
        return [fact.getPredicate().get_name() for fact in predicate_facts]

    @staticmethod
    def positive_predicate_proposition_facts(facts):
        for fact in facts:
            if fact.is_predicate_proposition() and fact.is_positive():
                yield fact



from typing import Text  # noqa: F401

from tala.utils.as_json import AsJSONMixin


class InputHypothesis(AsJSONMixin):
    def __init__(self, utterance, confidence):
        # type: (Text, float) -> None
        self._utterance = utterance
        self._confidence = confidence

    def as_dict(self):
        return {
            "utterance": self.utterance,
            "confidence": self.confidence,
        }

    @property
    def utterance(self):
        # type: () -> Text
        return self._utterance

    @property
    def confidence(self):
        # type: () -> float
        return self._confidence

    def __eq__(self, other):
        return self.utterance == other.utterance and self.confidence == other.confidence

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "InputHypothesis(%r, %r)" % (self._utterance, self._confidence)

    def __repr__(self):
        return str(self)



from contextlib import contextmanager
import locale

import iso8601


@contextmanager
def setlocale(name):
    previous = locale.setlocale(locale.LC_ALL)
    try:
        yield locale.setlocale(locale.LC_ALL, name)
    finally:
        locale.setlocale(locale.LC_ALL, previous)


class DateTime(object):
    def __init__(self, iso8601_string):
        self._iso8601_string = iso8601_string

    @property
    def iso8601_string(self):
        return self._iso8601_string

    def __eq__(self, other):
        try:
            return other.iso8601_string == self.iso8601_string
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (other == self)

    def __str__(self):
        return "datetime(%s)" % self.iso8601_string

    def human_standard(self, locale="en_US.UTF-8"):
        if locale != "en_US.UTF-8":
            raise UnsupportedLocale("Expected a supported locale for datetime but got '%s'" % locale)
        datetime = iso8601.parse_date(self._iso8601_string)
        with setlocale(locale):
            human_standard_string = datetime.strftime("%m/%d/%Y %I:%M %p", )
        return human_standard_string

    def __repr__(self):
        return 'DateTime("%s")' % self._iso8601_string

    def __hash__(self):
        return 17 * hash(str(self)) + 7 * hash(self.__class__.__name__)


class UnsupportedLocale(Exception):
    pass



import unittest

from tala.model.speaker import Speaker


class speakerTests(unittest.TestCase):
    def testSpeakerClass(self):
        self.assertEqual("SYS", Speaker.SYS)
        self.assertEqual("USR", Speaker.USR)



from tala.model.semantic_object import SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.unicodify import unicodify


class Hint(SemanticObjectWithContent, AsSemanticExpressionMixin):
    def __init__(self, plan):
        SemanticObjectWithContent.__init__(self, plan)
        self._plan = plan

    def __eq__(self, other):
        try:
            equality = self.plan == other.plan
            return equality
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self._plan)

    @property
    def plan(self):
        return self._plan

    def __str__(self):
        return "Hint" + unicodify(self.plan)



from mock import Mock

from tala.model.individual import Individual
from tala.model.ontology import Ontology, OntologyError
from tala.model.predicate import Predicate
from tala.model.sort import RealSort, IntegerSort, StringSort, CustomSort, ImageSort, WebviewSort, DateTimeSort, \
    PersonNameSort, Sort
from tala.model.image import Image
from tala.model.webview import Webview
from tala.model.date_time import DateTime
from tala.testing.lib_test_case import LibTestCase


class IndividualTestBase(LibTestCase):
    def setUp(self):
        self._create_ontology()
        self._create_semantic_objects()

    def _create_ontology(self):
        self.ontology_name = "mockup_ontology"
        self._city_sort = CustomSort(self.ontology_name, "city")
        sorts = {
            self._city_sort,
        }
        predicates = {
            self._create_predicate("price", RealSort()),
            self._create_predicate("number_of_passengers", IntegerSort()),
            self._create_predicate("map_to_show", ImageSort()),
            self._create_predicate("map_to_browse", WebviewSort()),
            self._create_predicate("departure_time", DateTimeSort()),
            self._create_predicate("comment", StringSort()),
            self._create_predicate("selected_city", self._city_sort),
            self._create_predicate("contact_to_call", PersonNameSort()),
        }
        individuals = {
            "paris": CustomSort(self.ontology_name, "city"),
            "london": CustomSort(self.ontology_name, "city"),
        }
        actions = set()
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

        self.empty_ontology = Ontology("empty_ontology", {}, {}, {}, set())

    def _create_predicate(self, *args, **kwargs):
        return Predicate(self.ontology_name, *args, **kwargs)

    def _create_semantic_objects(self):
        self.sort_city = self.ontology.get_sort("city")
        self.individual_paris = self.ontology.create_individual("paris")
        self.individual_not_paris = self.ontology.create_negative_individual("paris")
        self.integer_individual = self.ontology.create_individual(1234, IntegerSort())
        self.string_individual = self.ontology.create_individual("a string")
        self.image_individual = self.ontology.create_individual(Image("http://mymap.com/map.png"))
        self.webview_individual = self.ontology.create_individual(Webview("http://mymap.com/map.html"))
        self.datetime_individual = self.ontology.create_individual(DateTime("2018-04-11T22:00:00.000Z"))


class IndividualTest(IndividualTestBase):
    def test_create_real_individual_with_ontology_with_real_type(self):
        self.ontology.create_individual(123.50)

    def test_create_real_individual_with_ontology_without_real_type(self):
        ontology_without_real_type = self.empty_ontology
        with self.assertRaises(OntologyError):
            ontology_without_real_type.create_individual(123.50)

    def test_create_integer_individual_with_ontology_with_integer_type(self):
        self.ontology.create_individual(123)

    def test_create_integer_individual_with_ontology_without_integer_type(self):
        with self.assertRaises(OntologyError):
            ontology_without_integer_type = self.empty_ontology
            ontology_without_integer_type.create_individual(123)

    def test_real_individual_unicode(self):
        real_individual = self.ontology.create_individual(1234.0)
        self.assertEqual("1234.0", str(real_individual))

    def test_integer_individual_unicode(self):
        self.assertEqual("1234", str(self.integer_individual))

    def test_string_individual_unicode(self):
        self.assertEqual('"a string"', str(self.string_individual))

    def test_image_individual_unicode(self):
        self.assertEqual('image("http://mymap.com/map.png")', str(self.image_individual))

    def test_webview_individual_unicode(self):
        self.assertEqual('webview("http://mymap.com/map.html")', str(self.webview_individual))

    def test_datetime_individual_unicode(self):
        self.assertEqual('datetime(2018-04-11T22:00:00.000Z)', str(self.datetime_individual))

    def test_create_individual(self):
        individual = self.ontology.create_individual("paris")
        self.assertEqual("paris", individual.getValue())

    def test_create_negative_individual(self):
        individual = self.ontology.create_negative_individual("paris")
        self.assertEqual("paris", individual.getValue())

    def test_create_invalid_individual(self):
        ontology_without_string_predicate = self.empty_ontology
        with (self.assertRaises(OntologyError)):
            ontology_without_string_predicate.create_individual("kalle")

    def test_create_invalid_negative_individual(self):
        ontology_without_string_predicate = self.empty_ontology
        with (self.assertRaises(OntologyError)):
            ontology_without_string_predicate.create_negative_individual("kalle")

    def test_get_sort(self):
        individual = self.ontology.create_individual("paris")
        self.assertEqual(self.sort_city, individual.getSort())

    def test_equality(self):
        individual1 = self.ontology.create_individual("paris")
        individual2 = self.ontology.create_individual("paris")
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(individual1, individual2)

    def test_non_equality_due_to_value(self):
        other_individual = self.ontology.create_individual("london")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.individual_paris, other_individual)

    def test_non_equality_due_to_sort(self):
        other_individual = self.ontology.create_individual("paris", sort=StringSort())
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.individual_paris, other_individual)

    def test_non_equality_due_to_polarity(self):
        positive_individual = self.ontology.create_individual("paris")
        negative_individual = self.ontology.create_negative_individual("paris")
        self.assertNotEqual(positive_individual, negative_individual)
        self.assertNotEqual(negative_individual, positive_individual)

    def test_non_equality_with_non_individual(self):
        non_individual = "paris"
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.individual_paris, non_individual)

    def test_unicode_for_individual_of_custom_sort(self):
        self.assertEqual("paris", str(self.individual_paris))

    def test_unicode_for_negative_individual_of_custom_sort(self):
        self.assertEqual("~paris", str(self.individual_not_paris))

    def test_individual_is_positive(self):
        self.assertTrue(self.individual_paris.is_positive())

    def test_negative_individual_is_not_positive(self):
        self.assertFalse(self.individual_not_paris.is_positive())

    def test_individual_is_hashable(self):
        set_ = set()
        set_.add(self.individual_paris)

    def test_negative_individual_is_hashable(self):
        set_ = set()
        set_.add(self.individual_not_paris)

    def test_negate_positive_individual(self):
        self.assertFalse(self.individual_paris.negate().is_positive())

    def test_negate_negative_individual(self):
        self.assertTrue(self.individual_not_paris.negate().is_positive())

    def test_value_as_json_object(self):
        self.given_an_individual_of_mock_sort("mock_value")
        self.given_value_as_json_object_in_mock_sort_returns("mock_value", "mock_value_as_json_object")
        self.when_call_value_as_json_object_for_individual()
        self.then_result_is("mock_value_as_json_object")

    def given_an_individual_of_mock_sort(self, value):
        self._mock_sort = Mock(spec=Sort)
        self._individual = Individual("mock_ontology", value, self._mock_sort)

    def given_value_as_json_object_in_mock_sort_returns(self, expected_value, value_as_json_object):
        def mock_value_as_json_object(value):
            self.assertEqual(expected_value, value)
            return value_as_json_object
        self._mock_sort.value_as_json_object.side_effect = mock_value_as_json_object

    def when_call_value_as_json_object_for_individual(self):
        self._actual_result = self._individual.value_as_json_object()


class IsIndividualTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_individual_false(self):
        self.assertFalse(self.proposition_dest_city_paris.is_individual())

    def test_is_individual_true(self):
        self.assertTrue(self.individual_paris.is_individual())



import collections

from tala.utils.as_json import AsJSONMixin

SEMANTIC_OBJECT_TYPE = "ask_feature"


class AskFeature(collections.namedtuple("AskFeature", ["name", "kpq"]), AsJSONMixin):
    def __new__(cls, predicate_name, kpq=False):
        return super(AskFeature, cls).__new__(cls, predicate_name, kpq)

    def as_dict(self):
        return {"semantic_object_type": SEMANTIC_OBJECT_TYPE, "name": self.name, "kpq": self.kpq}



class BaseInteractionTest(object):
    @property
    def filename(self):
        raise NotImplementedError("This property needs to be implemented in a subclass")

    @property
    def name(self):
        raise NotImplementedError("This property needs to be implemented in a subclass")

    @property
    def turns(self):
        raise NotImplementedError("This property needs to be implemented in a subclass")



import pytest

from tala.model.ontology import Ontology, OntologyError, InvalidIndividualName, AmbiguousNamesException, IndividualExistsException, SortDoesNotExistException
from tala.model.predicate import Predicate
from tala.model.sort import RealSort, IntegerSort, ImageSort, DomainSort, CustomSort, BooleanSort, DateTimeSort, WebviewSort, StringSort, PersonNameSort, REAL, INTEGER, IMAGE, BOOLEAN, DATETIME, WEBVIEW, STRING
from tala.model.image import Image
from tala.model.webview import Webview
from tala.model.date_time import DateTime
from tala.model.person_name import PersonName
from tala.ddd.json_parser import JSONOntologyParser


class TestOntology(object):
    DEFAULT_NAME = "mockup_ontology"

    def _given_ontology(self, **kwargs):
        self._create_ontology(**kwargs)
        ontology_as_json = self._ontology.as_json()
        json_parser = JSONOntologyParser()
        self._ontology = json_parser.parse_ontology(ontology_as_json)

    def _when_creating_ontology(self, **kwargs):
        self._create_ontology(**kwargs)

    def _create_ontology(self, name=None, sorts=None, predicates=None, individuals=None, actions=None):
        if name is None:
            name = "mockup_ontology"
        if sorts is None:
            sorts = set()
        if predicates is None:
            predicates = set()
        if individuals is None:
            individuals = {}
        if actions is None:
            actions = []
        self._ontology = Ontology(name, sorts, predicates, individuals, actions)

    def _create_predicate(self, name, sort, *args, **kwargs):
        ontology_name = sort.ontology_name if sort.is_ontology_specific() else self.DEFAULT_NAME
        return Predicate(ontology_name, name, sort, *args, **kwargs)


class TestOntologyBasic(TestOntology):
    def test_get_name(self):
        self._given_ontology(name="mockup_ontology")
        assert "mockup_ontology" == self._ontology.get_name()

    def test_sorts_base_case(self):
        self._given_ontology(name=self.DEFAULT_NAME, sorts={CustomSort(self.DEFAULT_NAME, "city")})
        self._when_get_sorts_is_called()
        self._then_result_contains_key("city", CustomSort(self._ontology.name, "city"))

    def _when_get_sorts_is_called(self):
        self._result = self._ontology.get_sorts()

    def _then_result_contains_key(self, key, value):
        assert value == self._result[key]

    def test_predicates_base_case(self):
        self._given_ontology(
            name=self.DEFAULT_NAME,
            predicates={self._create_predicate("dest_city", CustomSort(self.DEFAULT_NAME, "city"))},
            sorts={CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_get_predicates_is_called()
        self._then_result_is({
            "dest_city": self._create_predicate("dest_city", CustomSort(self._ontology.name, "city"))
        })

    def _when_get_predicates_is_called(self):
        self._result = self._ontology.get_predicates()

    def test_individuals_base_case(self):
        self._given_ontology(
            name=self.DEFAULT_NAME,
            individuals={"paris": CustomSort(self.DEFAULT_NAME, "city")},
            sorts={CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_get_individuals_is_called()
        self._then_result_is({"paris": CustomSort(self.DEFAULT_NAME, "city")})

    def _when_get_individuals_is_called(self):
        self._result = self._ontology.get_individuals()

    def test_actions_with_additional_action(self):
        self._given_ontology(actions={"buy"})
        self._when_get_actions_is_called()
        self._then_result_is({"buy", "top", "up", "how"})

    def _when_get_actions_is_called(self):
        self._result = self._ontology.get_actions()

    def _then_result_contains(self, obj):
        assert obj in self._result

    def test_get_ddd_specific_actions_with_additional_action(self):
        self._given_ontology(actions={"buy"})
        self._when_get_ddd_specific_actions_is_called()
        self._then_result_is({"buy"})

    def test_default_sorts(self):
        self._given_ontology()
        self._when_get_sorts_is_called()
        self._then_result_contains_key("domain", DomainSort())

    def test_builtin_sorts_added_for_predicates(self):
        self._given_ontology(name=self.DEFAULT_NAME, predicates={self._create_predicate("price", RealSort())})
        self._when_get_sorts_is_called()
        self._then_result_contains_key("real", RealSort())

    def test_default_actions_included_in_get_actions(self):
        self._given_ontology()
        self._when_get_actions_is_called()
        self._then_result_is({"top", "up", "how"})

    def test_default_actions_missing_from_get_ddd_specific_actions(self):
        self._given_ontology()
        self._when_get_ddd_specific_actions_is_called()
        self._then_result_is(set())

    def _when_get_ddd_specific_actions_is_called(self):
        self._result = self._ontology.get_ddd_specific_actions()

    def test_individual_sort_for_custom_sort(self):
        self._given_ontology(
            name=self.DEFAULT_NAME,
            sorts={CustomSort(self.DEFAULT_NAME, "city")},
            predicates={self._create_predicate("selected_city", CustomSort(self.DEFAULT_NAME, "city"))},
            individuals={"paris": CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_individual_sort_is_called("paris")
        self._then_result_is(CustomSort(self.DEFAULT_NAME, "city"))

    def _when_individual_sort_is_called(self, name):
        self._result = self._ontology.individual_sort(name)

    def _then_result_is(self, expected_result):
        assert expected_result == self._result

    def test_individual_sort_for_builtin_sort(self):
        self._given_ontology(predicates={self._create_predicate("price", RealSort())})
        self._when_individual_sort_is_called(123.50)
        self._then_result_is(RealSort())

    def test_individual_sort_for_undefined_sort_raises_exception(self):
        self._given_ontology()
        with pytest.raises(OntologyError):
            self._when_individual_sort_is_called("kalle")

    def test_individual_sort_for_boolean_supported_by_ontology(self):
        self._given_ontology(predicates={self._create_predicate("needs_visa", BooleanSort())})
        self._when_individual_sort_is_called(True)
        self._then_result_is(BooleanSort())

    def test_individual_sort_for_boolean_unsupported_by_ontology(self):
        self._given_ontology()
        self._when_individual_sort_is_called_then_exception_is_raised_matching(
            True, OntologyError, "Expected one of the predicate sorts [] in ontology '{}', "
            "but got value 'True' of sort 'boolean'".format(self.DEFAULT_NAME)
        )

    def _when_individual_sort_is_called_then_exception_is_raised_matching(
        self, value, expected_exception, expected_message
    ):
        with pytest.raises(expected_exception) as exception_information:
            self._ontology.individual_sort(value)
        assert expected_message == str(exception_information.value)

    def test_individual_sort_for_datetime_supported_by_ontology(self):
        self._given_ontology(predicates={self._create_predicate("departure_time", DateTimeSort())})
        self._when_individual_sort_is_called(DateTime("2018-03-20T22:00:00.000Z"))
        self._then_result_is(DateTimeSort())

    def test_individual_sort_for_datetime_unsupported_by_ontology(self):
        self._given_ontology()
        self._when_individual_sort_is_called_then_exception_is_raised_matching(
            DateTime("2018-03-20T22:00:00.000Z"), OntologyError,
            "Expected one of the predicate sorts [] in ontology '{}', "
            "but got value 'datetime(2018-03-20T22:00:00.000Z)' of sort 'datetime'".format(self.DEFAULT_NAME)
        )

    def test_is_action_true(self):
        self._given_ontology(actions={"buy"})
        self._when_is_action_is_called("buy")
        self._then_result_is(True)

    def _when_is_action_is_called(self, obj):
        self._result = self._ontology.is_action(obj)

    def test_is_action_false(self):
        self._given_ontology(actions={"buy"})
        self._when_is_action_is_called("destination")
        self._then_result_is(False)

    def test_predicate_declaration_with_unknown_sort_yields_exception(self):
        with pytest.raises(OntologyError):
            self._when_creating_ontology(
                sorts=set(),
                predicates={self._create_predicate("some_predicate", CustomSort(self.DEFAULT_NAME, "unknown_sort"))},
                individuals={},
                actions=set(),
                name="InvalidOntology"
            )

    def test_individual_declaration_with_unknown_sort_yields_exception(self):
        with pytest.raises(OntologyError):
            self._when_creating_ontology(
                sorts=set(),
                predicates=set(),
                individuals={"some_individual": CustomSort(self.DEFAULT_NAME, "unknown_sort")},
                actions=set(),
                name="InvalidOntology"
            )

    def test_has_predicate_true(self):
        self._given_ontology(
            predicates={self._create_predicate("dest_city", CustomSort(self.DEFAULT_NAME, "city"))},
            sorts={CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_has_predicate_is_called("dest_city")
        self._then_result_is(True)

    def _when_has_predicate_is_called(self, obj):
        self._result = self._ontology.has_predicate(obj)

    def test_has_predicate_false(self):
        self._given_ontology()
        self._when_has_predicate_is_called("dest_city")
        self._then_result_is(False)

    def test_has_sort_true(self):
        self._given_ontology(sorts={CustomSort(self.DEFAULT_NAME, "city")})
        self._when_has_sort_is_called("city")
        self._then_result_is(True)

    def _when_has_sort_is_called(self, obj):
        self._result = self._ontology.has_sort(obj)

    def test_has_sort_false(self):
        self._given_ontology()
        self._when_has_sort_is_called("city")
        self._then_result_is(False)

    def test_invalid_individual_name_in_constructor_yields_exception(self):
        with pytest.raises(InvalidIndividualName):
            self._when_creating_ontology(individuals={"name with whitespace": IntegerSort()})

    def test_action_shares_name_with_predicate_raises_error(self):
        with pytest.raises(AmbiguousNamesException):
            self._when_creating_ontology(
                sorts=set(),
                predicates={self._create_predicate("price", RealSort())},
                individuals={},
                actions={"price"}
            )

    def test_is_individual_static(self):
        city_sort = CustomSort(self.DEFAULT_NAME, "city")
        self._when_creating_ontology(sorts={city_sort}, individuals={"paris": city_sort})
        self._then_individual_is_static("paris")

    def _then_individual_is_static(self, name):
        assert self._ontology.is_individual_static(name) is True

    @pytest.mark.parametrize(
        "sort,name", [(RealSort(), REAL), (BooleanSort(), BOOLEAN), (IntegerSort(), INTEGER), (ImageSort(), IMAGE),
                      (WebviewSort(), WEBVIEW), (StringSort(), STRING), (DateTimeSort(), DATETIME)]
    )
    def test_predicates_contain_builtin_sort_base_case(self, sort, name):
        self._given_ontology(predicates={self._create_predicate("predicate", sort)}, )
        self._when_predicates_contain_sort_is_called_with(name)
        self._then_result_is(True)

    @pytest.mark.parametrize(
        "sort,name", [(RealSort(), REAL), (BooleanSort(), BOOLEAN), (IntegerSort(), INTEGER), (ImageSort(), IMAGE),
                      (WebviewSort(), WEBVIEW), (StringSort(), STRING), (DateTimeSort(), DATETIME)]
    )
    def test_predicates_contain_builtin_sort_when_ontology_has_predicates_of_custom_sorts(self, sort, name):
        self._given_ontology(
            sorts={CustomSort(self.DEFAULT_NAME, "sort"),
                   CustomSort(self.DEFAULT_NAME, "another sort")},
            predicates={
                self._create_predicate("first predicate", CustomSort(self.DEFAULT_NAME, "sort")),
                self._create_predicate("second predicate", sort),
                self._create_predicate("third predicate", CustomSort(self.DEFAULT_NAME, "another sort"))
            },
        )
        self._when_predicates_contain_sort_is_called_with(name)
        self._then_result_is(True)

    def _when_predicates_contain_sort_is_called_with(self, sort):
        self._result = self._ontology.predicates_contain_sort(sort)

    @pytest.mark.parametrize("name", [REAL, BOOLEAN, INTEGER, IMAGE, WEBVIEW, STRING, DATETIME])
    def test_predicates_do_not_contain_builtin_sort_base_case(self, name):
        self._given_ontology(name=self.DEFAULT_NAME)
        self._when_predicates_contain_sort_is_called_with(name)
        self._then_result_is(False)

    @pytest.mark.parametrize("name", [REAL, BOOLEAN, INTEGER, IMAGE, WEBVIEW, STRING, DATETIME])
    def test_predicates_do_not_contain_builtin_sort_when_ontology_only_contains_predicate_of_custom_sort(self, name):
        self._given_ontology(
            name=self.DEFAULT_NAME,
            sorts={CustomSort(self.DEFAULT_NAME, "city")},
            predicates={self._create_predicate("selected_city", CustomSort(self.DEFAULT_NAME, "city"))},
            individuals={"paris": CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_predicates_contain_sort_is_called_with(name)
        self._then_result_is(False)

    @pytest.mark.parametrize(
        "value,sort", [(25.5, REAL), (True, BOOLEAN), (14, INTEGER), (Image("image-url"), IMAGE),
                       (Webview("webview-url"), WEBVIEW), ("a string", STRING),
                       (DateTime("2018-03-20T22:00:00.000Z"), DATETIME)]
    )
    def test_individual_sort_for_builtin_sort_unsupported_by_ontology_base_case(self, value, sort):
        self._given_ontology(name=self.DEFAULT_NAME)
        self._when_individual_sort_is_called_then_exception_is_raised_matching(
            value, OntologyError, "Expected one of the predicate sorts [] in ontology '{}', "
            "but got value '{}' of sort '{}'".format(self.DEFAULT_NAME, value, sort)
        )

    @pytest.mark.parametrize(
        "value,sort", [(25.5, REAL), (True, BOOLEAN), (14, INTEGER), (Image("image-url"), IMAGE),
                       (Webview("webview-url"), WEBVIEW), ("a string", STRING),
                       (DateTime("2018-03-20T22:00:00.000Z"), DATETIME)]
    )
    def test_individual_sort_for_builtin_sort_unsupported_by_ontology_when_ontology_has_custom_sorts(self, value, sort):
        self._given_ontology(
            name=self.DEFAULT_NAME,
            sorts={CustomSort(self.DEFAULT_NAME, "a sort"),
                   CustomSort(self.DEFAULT_NAME, "another sort")},
            predicates={
                self._create_predicate("a predicate 1", CustomSort(self.DEFAULT_NAME, "a sort")),
                self._create_predicate("a predicate 2", CustomSort(self.DEFAULT_NAME, "a sort")),
                self._create_predicate("another predicate 1", CustomSort(self.DEFAULT_NAME, "another sort")),
                self._create_predicate("another predicate 2", CustomSort(self.DEFAULT_NAME, "another sort"))
            },
            individuals={"individual": CustomSort(self.DEFAULT_NAME, "a sort")}
        )
        self._when_individual_sort_is_called_then_exception_is_raised_matching(
            value, OntologyError, "Expected one of the predicate sorts ['a sort', 'another sort'] in ontology '{}', "
            "but got value '{}' of sort '{}'".format(self.DEFAULT_NAME, value, sort)
        )

    def test_identical_sort_and_predicate_name_raises_exception(self):
        self._when_creating_ontology_then_exception_is_raised(
            sorts={CustomSort(self.DEFAULT_NAME, "mock-name")},
            predicates={self._create_predicate("mock-name", CustomSort(self.DEFAULT_NAME, "mock-name"))},
            expected_exception=AmbiguousNamesException,
            expected_message="Expected predicate and sort names to be unique but there is both a predicate and sort "
                             f"named 'mock-name' in ontology '{self.DEFAULT_NAME}'"
        )  # yapf: disable

    def _when_creating_ontology_then_exception_is_raised(self, expected_exception, expected_message, *args, **kwargs):
        with pytest.raises(expected_exception, match=expected_message):
            self._create_ontology(*args, **kwargs)


class TestDynamicOntology(TestOntology):
    def test_add_dynamic_individual(self):
        self._given_ontology(sorts={CustomSort(self.DEFAULT_NAME, "city")})
        self._when_individual_is_added("paris", "city")
        self._then_individual_is_in_the_ontology("paris", "city")

    def _when_individual_is_added(self, name, sort):
        self._ontology.add_individual(name, sort)

    def _then_individual_is_in_the_ontology(self, name, sort):
        sort_as_instance = self._ontology.get_sort(sort)
        assert (name, sort_as_instance) in list(self._ontology.get_individuals().items())

    def test_adding_existing_individual_yields_exception(self):
        self._given_ontology(
            sorts={CustomSort(self.DEFAULT_NAME, "city")}, individuals={"paris": CustomSort(self.DEFAULT_NAME, "city")}
        )
        with pytest.raises(IndividualExistsException):
            self._when_individual_is_added("paris", "city")

    def test_add_individual_with_non_existing_sort_yields_exception(self):
        self._given_ontology()
        with pytest.raises(SortDoesNotExistException):
            self._when_individual_is_added("athens", "river")

    def test_ensure_individual_exists_non_existing_individual(self):
        self._given_ontology(sorts={CustomSort(self.DEFAULT_NAME, "city")})
        self._when_ensure_individual_exists_is_called("paris", "city")
        self._then_individual_is_in_the_ontology("paris", "city")

    def _when_ensure_individual_exists_is_called(self, name, sort):
        self._ontology.ensure_individual_exists(name, sort)

    def test_ensure_individual_exists_existing_individual(self):
        self._given_ontology(
            sorts={CustomSort(self.DEFAULT_NAME, "city")}, individuals={"paris": CustomSort(self.DEFAULT_NAME, "city")}
        )
        self._when_ensure_individual_exists_is_called("paris", "city")
        self._then_no_exception_is_read()

    def _then_no_exception_is_read(self):
        pass

    def test_add_individual_checks_name(self):
        self._given_ontology()
        with pytest.raises(InvalidIndividualName):
            self._when_individual_is_added("name with whitespace", "integer")

    def test_ensure_individual_exists_checks_name(self):
        self._given_ontology()
        with pytest.raises(InvalidIndividualName):
            self._when_ensure_individual_exists_is_called("name with whitespace", "integer")

    def test_adding_dynamic_individual_is_not_considered_static(self):
        self._given_ontology(sorts={CustomSort(self.DEFAULT_NAME, "city")})
        self._when_individual_is_added("paris", "city")
        self._then_individual_is_not_static("paris")

    def _then_individual_is_not_static(self, name):
        assert self._ontology.is_individual_static(name) is False

    def test_create_person_name_individual(self):
        self._given_ontology(sorts={})
        self._when_individual_is_created("Anna Kronlid", PersonNameSort())
        self._then_individual_is_of_right_type()

    def _when_individual_is_created(self, individual_string, sort):
        self._individual = self._ontology.create_individual(individual_string, sort)

    def _then_individual_is_of_right_type(self):
        assert self._individual.getSort() == PersonNameSort()
        assert self._individual.getValue() == PersonName("Anna Kronlid")



from tala.testing.interactions.base_test import BaseInteractionTest


class InteractionTest(BaseInteractionTest):
    def __init__(self, file_name, test_name, turns):
        self._file_name = file_name
        self._test_name = test_name
        self._turns = turns

    @property
    def filename(self):
        return self._file_name

    @property
    def name(self):
        return self._test_name

    @property
    def turns(self):
        return self._turns



#!/usr/bin/env python3.9

from setuptools import find_packages, setup

try:
    from setuptools.command.build_py import build_py
except ImportError:
    from distutils.command.build_py import build_py  # noqa: F401


def readme():
    with open("README.md") as f:
        return f.read()


if __name__ == "__main__":
    setup(
        use_scm_version={
            "write_to": "tala/installed_version.py",
            "root": ".",
            "relative_to": __file__,
            "git_describe_command": "git describe --dirty --tags --long "
                    "--exclude tdm-* --exclude *candidate* --exclude *latest* --exclude *master* ",
        },
        setup_requires=["setuptools_scm"],
        name="tala",
        description="Design dialogue domain descriptions (DDDs) for TDM",
        long_description=readme(),
        long_description_content_type="text/markdown",
        classifiers=[
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.9",
        ],
        keywords="tala tdm ddd ddds dialogue conversation AI",
        packages=find_packages(exclude=["tala/ddds", "test", "*.test", "test.*", "*.test.*"]),
        package_dir={"tala": "tala"},
        package_data={
            "tala": [
                "ddd/maker/templates/*.txt",
                "ddd/maker/templates/*.xml",
                "ddd/schemas/grammar.xsd",
                "ddd/schemas/grammar_rgl.xsd",
                "ddd/schemas/ontology.xsd",
                "ddd/schemas/service_interface.xsd",
                "ddd/schemas/domain.xsd",
                "nl/entity_examples/*/*.csv",
            ]
        },
        scripts=[],
        entry_points={
            "console_scripts": [
                "tala = tala.cli.console_script:main",
            ],
        },
        url="http://www.talkamatic.se",
        author="Talkamatic",
        author_email="dev@talkamatic.se",
        install_requires=[
            "Jinja2>=3.0.1, <4",
            "dill>=0.3.6",
            "iso8601>=0.1.14",
            "lxml>=4.6.3",
            "prompt-toolkit>=3.0.19",
            "requests>=2.26.0",
            "structlog>=21.1.0",
            "setuptools>=68.2.2",
            "setuptools-scm>=8.0.4"

        ],
        dependency_links=[],
    )  # yapf: disable
